import fnmatch
import signal
import threading
import webbrowser

import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_table
import numpy
from dash.dependencies import Input, Output
from dash_table.Format import Format, Scheme
from flask import Flask
from gevent.pywsgi import WSGIServer

from .utils import *

server_address = ("localhost", 8050)
server = None
appserver = None


class WebServer(threading.Thread):
    def __init__(self):
        super().__init__()

    def run(self):
        global server
        server = WSGIServer(server_address, appserver, log=None)
        server.serve_forever()


def shutdown(num, info):
    print(f'Shutting down server on port {server_address[1]}.\n')
    server.stop()
    server.close()


def main(config_file):
    df = df_setup(config_file)
    dash_setup(df)

    WebServer().start()
    webbrowser.open_new('http://127.0.0.1:8050/')
    signal.signal(signal.SIGINT, shutdown)

    print('Use CTRL+C to close the server.')


def df_setup(config_file):
    print(f'--- Creating interactive report on server localhost:{server_address[1]} ---')
    print('Select the results directory created by fRAT.')
    folder = Utils.file_browser(title='Select the directory output by the fRAT')

    config = Utils.load_config(f'{Path(os.path.abspath(__file__)).parents[1]}/configuration_profiles/', config_file)

    if config.averaging_type == 'Session averaged':
        subfolder = 'Session_averaged_results'
        print(f"Showing {subfolder.replace('_', ' ')}.")

    else:
        subfolder = 'Participant_averaged_results'
        print(f"Showing {subfolder.replace('_', ' ')}.")

    df = pd.read_json(f"{folder}/Overall/Summarised_results/{subfolder}/combined_results.json")

    r = re.compile("Conf_Int_*")
    Conf_Int = list(filter(r.match, df.columns))[0]  # Find confidence interval level

    if config.averaging_type == 'Session averaged':
        column_order = ['index', 'Mean', Conf_Int, 'Median', 'Std_dev',
                        'Minimum', 'Maximum', 'Total voxels', 'Excluded voxels', 'Average voxels per session',
                        'Sessions', 'File_name']
    else:
        column_order = ['index', 'Mean', Conf_Int, 'Median', 'Std_dev',
                        'Minimum', 'Maximum', 'Total voxels',
                        'Excluded voxels', 'Average voxels per session',
                        'Participants', 'Sessions', 'File_name']

    if 'Percentage change from baseline' in df.columns:
        column_order.insert(3, 'Percentage change from baseline')
        column_order.insert(4, 'Baseline')

    crit_params = list(set(df.columns) ^ set(column_order))  # Find elements not in both lists

    for crit_param in crit_params:
        column_order.insert(1, crit_param)  # Insert critical parameters after index

    df = df[column_order]  # Reorganise columns

    df["id"] = df.index

    return df


def dash_setup(df):
    global appserver

    colors = {'background': '#f8f8f8', 'text': '#000000'}
    external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

    appserver = Flask(__name__)
    app = dash.Dash(__name__, server=appserver, external_stylesheets=external_stylesheets)

    columns = []
    for i, key in enumerate(df.columns):
        # Column format setup
        columns.append({"name": key, "id": key, "deletable": True, "selectable": True})

        if isinstance(df[key][0], numpy.float64):
            columns[i].update({"type": "numeric", "format": Format(precision=2, scheme=Scheme.fixed)})
        elif isinstance(df[key][0], numpy.int64):
            columns[i].update({"type": "numeric", "format": Format(precision=2, scheme=Scheme.decimal_integer)})

    app.layout = html.Div(style={'backgroundColor': colors['background']}, children=[
        dcc.Markdown('# fRAT Report \n Interactive table', style={'textAlign': 'center', 'color': colors['text']}),
        dash_table.DataTable(
            id='datatable',
            columns=columns,
            data=df.to_dict('records'),
            filter_action="native",
            sort_action="native",
            sort_mode="multi",
            row_deletable=True,
            row_selectable='multi',
            selected_rows=[],
            page_action="native",
            page_current=0,
            page_size=len(df['index'].unique()),
            style_cell={'textAlign': 'left'},
        ),

        html.Div([html.Label(['Statistic to display:',

                              dcc.Dropdown(id='barchart_dropdown',
                                           options=[
                                               {'label': 'Mean', 'value': 'Mean'},
                                               {'label': 'Standard deviation', 'value': 'Std_dev'},
                                               {'label': 'Median', 'value': 'Median'},
                                               {'label': 'Minimum', 'value': 'Minimum'},
                                               {'label': 'Maximum', 'value': 'Maximum'},
                                               {'label': 'Total voxels', 'value': 'Total voxels'},
                                               {'label': 'Average voxels per session', 'value': 'Average voxels per session'},
                                               {'label': 'Excluded voxels (percentage)', 'value': 'Excluded voxels'}
                                           ],
                                           value='Mean',
                                           style={"width": "40%"},
                                           clearable=False
                                           )]),
                  html.Br(),
                  dcc.Graph(id='barchart', figure={"layout": {"height": 500}})
                  ],
                 style={"width": "70%", 'marginLeft': 40, 'marginRight': 10, 'marginTop': 10, 'marginBottom': 10,
                        'padding': '10px 0px 0px 10px'},
                 )
    ])

    @app.callback(Output("datatable", "style_data_conditional"),
                  Input("datatable", "derived_viewport_selected_row_ids"))
    def style_selected_rows(selected_rows):
        if selected_rows is None:
            return dash.no_update

        style = [{'if': {'row_index': 'odd'}, 'backgroundColor': 'rgb(248, 248, 248)'}]
        [style.append({"if": {"filter_query": f"{{id}} ={i}"}, "backgroundColor": "#D2F3FF"}) for i in selected_rows]

        return style

    @app.callback(
        Output('barchart', 'figure'),
        [Input('datatable', 'selected_row_ids'),
         Input('barchart_dropdown', 'value'),
         Input('barchart_dropdown', 'options')])
    def update_graph(selected_rows, dropdown_value, dropdown_options):
        if selected_rows is None:
            return dash.no_update

        dff = df.loc[df['id'].isin(selected_rows)].copy()

        dropdown_label = [i['label'] for i in dropdown_options if i['value'] == dropdown_value][0]

        ConfInts = []
        if dropdown_value == 'Mean':
            ConfInts = dff[fnmatch.filter(df.columns, 'Conf_Int*')[0]]

        elif dropdown_value == 'Excluded voxels':
            dff['Excluded voxels'] = dff['Excluded voxels'] / (dff['Excluded voxels'] + dff['Total voxels']) * 100
            dropdown_label = 'Percentage of voxels excluded'

        intwidth = 35
        barwidth = 0.09 * (len(selected_rows))

        return {
            'data': [{"x": dff['index'],
                     "y": dff[dropdown_value],
                     "width": barwidth,
                     "error_y": {"type": 'data', "array": ConfInts, "thickness": 2, "width": intwidth},
                     "type": 'bar'}],

            "layout": {
                'title': 'Interactive barchart',
                "xaxis": {"title": "Region of interest"},
                "yaxis": {"title": dropdown_label},
                'plot_bgcolor': colors['background'],
                'paper_bgcolor': colors['background']
            },
        }



import copy
import itertools
import os
import random
import re
import shutil
import warnings
from glob import glob
from pathlib import Path

import nibabel as nib
import numpy as np
import numpy.linalg
import pandas as pd
import plotnine as pltn
import matplotlib.pyplot as plt
import statsmodels.api as sm
import statsmodels.formula.api as smf
import statsmodels.tools.sm_exceptions
from nilearn import plotting
from scipy.stats import zscore
from statsmodels.stats.weightstats import ttest_ind
from scipy.stats import ttest_rel
from statsmodels.tools.sm_exceptions import ConvergenceWarning

from .analysis import Environment_Setup, MatchedBrain
from .figures import Histogram, BrainGrid
from .utils import Utils

config = None
STATISTICS_PATH = ''
STATISTICS_LOGFILE = None


class Coefficient_map(MatchedBrain, Environment_Setup):
    @classmethod
    def label_list_setup(cls):
        cls.atlas_path = f'{os.environ["FSLDIR"]}/data/atlases/' \
                         f'{cls.atlas_label_list[int(config.atlas_number)][0]}'
        cls.atlas_label_path = f'{os.environ["FSLDIR"]}/data/atlases/' \
                               f'{cls.atlas_label_list[int(config.atlas_number)][1]}'
        cls.roi_label_list()

    @classmethod
    def to_braingrid(cls, df, subfolder, standardise_cbar, subsubfolder='/'):
        cls.label_list_setup()

        Utils.check_and_make_dir(f"{STATISTICS_PATH}/{subfolder}/")
        Utils.check_and_make_dir(f"{STATISTICS_PATH}/{subfolder}/{subsubfolder}/")
        Utils.check_and_make_dir(f"{STATISTICS_PATH}/{subfolder}/{subsubfolder}/NIFTI_ROI")
        Utils.check_and_make_dir(f"{STATISTICS_PATH}/{subfolder}/{subsubfolder}/Images")

        subfolder = f"{subfolder}/{subsubfolder}"

        cls.create_images(df, subfolder, standardise_cbar)

    @classmethod
    def save_brain_imgs(cls, file_name, subfolder, vmax):
        # Save brain image using nilearn
        warnings.filterwarnings("ignore", category=UserWarning)  # Turn off warning regarding nan values being converted to zeros
        warnings.filterwarnings("ignore", category=FutureWarning)
        plot = plotting.plot_stat_map(f"{STATISTICS_PATH}/{subfolder}/NIFTI_ROI/{file_name}.nii.gz",
                                      draw_cross=False, annotate=False, colorbar=True,
                                      vmax=vmax, symmetric_cbar=True, cbar_tick_format="%.2f",
                                      display_mode='xz', cut_coords=(0, 18),
                                      cmap='seismic')
        plot.savefig(f"{STATISTICS_PATH}/{subfolder}/Images/{file_name}.png")
        plot.close()
        warnings.filterwarnings("default", category=UserWarning)
        warnings.filterwarnings("default", category=FutureWarning)

    @classmethod
    def create_images(cls, df, subfolder, standardise_cbar):
        if standardise_cbar:
            # Find max value seen in df so the same colour bar can be used
            vmax = max(abs(np.min(df.iloc[:, 0])), abs(np.max(df.iloc[:, 0])))
        else:
            vmax = None

        for variable in df.index.unique():
            curr_df = df.loc[variable]

            try:
                data = [curr_df.iloc[:, 0].values]
                columns = curr_df['ROI'].values
            except pd.core.indexing.IndexingError:
                data = [curr_df.values[0]]
                columns = [curr_df[1]]

            reorganised_df = pd.concat([pd.DataFrame(columns=cls.label_array),
                                        pd.DataFrame(data=data,
                                                     columns=columns)])

            if variable[0:2] == 'C(':
                # Reformat filename if variable is categorical as it would be in the wrong format, e.g. C(MB)[T.On]
                variable = re.sub('\(|\)|\[|\]', '', variable)
                variable = re.sub('(T\.)', '_', variable)[1:]

            cls.scale_and_save_atlas_images(cls.atlas_path, None, reorganised_df.to_numpy(),
                                            0, f"{STATISTICS_PATH}/{subfolder}/NIFTI_ROI", variable)
            cls.save_brain_imgs(variable, subfolder, vmax)


def main(cfg, config_path, config_filename):
    global config

    config = cfg

    Coefficient_map.label_list_setup()

    stats_setup(config_path, config_filename)
    param_queries, critical_params = find_baseline_parameters()

    # Only use selected parameters for statistics
    critical_params = [param for param in critical_params if
                       param.lower() in [x.lower() for x in config.include_as_variable]]

    statistics(param_queries, critical_params)

    STATISTICS_LOGFILE.close()


def stats_setup(config_path, config_filename):
    global STATISTICS_PATH, STATISTICS_LOGFILE

    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    STATISTICS_PATH = f'{os.getcwd()}/Statistics/{config.statistics_subfolder_name}/'
    Utils.check_and_make_dir(f"{os.getcwd()}/Statistics")
    Utils.check_and_make_dir(STATISTICS_PATH)

    # Save copy of statistics options used
    shutil.copyfile(f"{os.getcwd()}/../statisticsOptions.csv", f"{STATISTICS_PATH}/copy_statisticsOptions.csv")

    # Save log of only the statistics settings from the config file
    Utils.save_config(STATISTICS_PATH, config_path, config_filename,
                      relevant_sections=['Statistics'], new_config_name='statistics_log',
                      additional_info=[
                          f'data_used_for_statistics = "{config.averaging_type.replace(" ", "_").lower()}"'])

    STATISTICS_LOGFILE = open(f'{os.getcwd()}/Statistics/{config.statistics_subfolder_name}/'
                              f'statistics_terminal_output.txt', 'w')

    Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                         f'\nCalculating statistics using {config.averaging_type} data.')


def clean_results(individual_roi_results, excluded_parameters, raw_data,
                  averaged_roi_results, combined_roi_results, voxel_cutoff):
    """For each ROI, remove each session that has a voxel count below the minimum voxel cutoff and make a log of which
    ROIs this applies to."""
    rois_below_max_r2_thresh = []

    Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                         f"\nNOTE: Within-subjects t-tests are paired Student's t-tests, however between-subjects "
                         f"t-tests are Welch's t-tests with degrees of freedom calculated using Satterthwaite Approximation."
                         f'\nMinimum voxel count set to {voxel_cutoff}.\n\n'
                         f'Sessions removed:')

    # Remove No ROI and overall data to be used with percentage change statistics
    combined_roi_results = combined_roi_results[(combined_roi_results['index'] != 'No ROI')
                                                & (combined_roi_results['index'] != 'Overall')]
    combined_roi_results = combined_roi_results[combined_roi_results['Average voxels per session'] > voxel_cutoff]

    for key in excluded_parameters:
        combined_roi_results = combined_roi_results.loc[(~combined_roi_results[key].isin(excluded_parameters[key]))]

    for roi in individual_roi_results:
        for key in excluded_parameters:
            individual_roi_results[roi] = individual_roi_results[roi].loc[
                (~individual_roi_results[roi][key].isin(excluded_parameters[key]))]

            raw_data[roi] = raw_data[roi].loc[(~raw_data[roi][key].isin(excluded_parameters[key]))]

            averaged_roi_results[roi] = averaged_roi_results[roi].loc[
                (~averaged_roi_results[roi][key].isin(excluded_parameters[key]))]

        session_count = len(individual_roi_results[roi])
        below_thresh_number = np.count_nonzero(individual_roi_results[roi]['voxel_amount'] < voxel_cutoff)

        individual_roi_results[roi] = individual_roi_results[roi][
            individual_roi_results[roi]['voxel_amount'] > voxel_cutoff]

        averaged_roi_results[roi] = averaged_roi_results[roi][averaged_roi_results[roi]['voxel_amount'] > voxel_cutoff]

        if below_thresh_number:
            percent_removed = (below_thresh_number / session_count) * 100

            if percent_removed > config.max_below_thresh:
                rois_below_max_r2_thresh.append(roi)

            Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                                 f'{roi}: {below_thresh_number}/{session_count} ({"{:.2f}".format(percent_removed)}%)')

    return rois_below_max_r2_thresh, individual_roi_results, averaged_roi_results, raw_data, combined_roi_results


def extract_statistics_options_info():
    statistics_options_table = pd.read_csv(f"{os.getcwd()}/../statisticsOptions.csv", header=None)
    statistics_options_table = np.split(statistics_options_table,
                                        statistics_options_table[statistics_options_table.isnull().all(1)].index)

    # Drop any blank rows
    statistics_options_table = [df.dropna(how='all') for df in statistics_options_table if
                                not df.dropna(how='all').empty]

    main_effect_parameters = {}
    simple_effect_parameters = {}
    excluded_parameters = {}

    for counter, parameter in enumerate(statistics_options_table):
        if counter % 2 == 0:
            main_effect_temp_list = parameter[0].loc[parameter[1].isin(['y', 'Y', 'yes', 'Yes'])]
            main_effect_parameters[parameter.values[0][0]] = list(main_effect_temp_list.values)

        else:
            excluded_values_temp_list = parameter[0].loc[parameter[2].isin(['y', 'Y', 'yes', 'Yes'])]
            excluded_parameters[parameter.values[0][0]] = list(excluded_values_temp_list.values)

            simple_effects_temp_list = parameter[0].loc[parameter[1].isin(['y', 'Y', 'yes', 'Yes'])]
            simple_effect_parameters[parameter.values[0][0]] = list(simple_effects_temp_list.values)

    return main_effect_parameters, simple_effect_parameters, excluded_parameters


def find_baseline_parameters():
    baseline_param_query = ''
    non_baseline_params_query = ''

    table, _ = Utils.load_paramValues_file()
    columns = table.columns
    _, critical_column_locs, baseline_column_loc = Utils.find_column_locs(table)

    baseline_params = []  # Make a list in case user has selected multiple rows
    for index, row in table.iterrows():
        if str(row[baseline_column_loc]).strip().lower() in ('yes', 'y', 'true'):
            baseline_params.append(dict(row[critical_column_locs]))

    # Check number of unique param combinations given, need to convert from dict to str as dict is not hashable
    if len(set([str(x) for x in baseline_params])) > 1:
        raise ValueError('Multiple parameter combinations have been selected as the baseline to use for statistics.\n'
                         'Only one row should be selected, the other brain volumes with the same parameter combination '
                         'will also be used for the baseline.')

    elif len(set([str(x) for x in baseline_params])) == 0:
        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f'\nNo baseline set in parameter values file. Skipping calculation of percent change '
                             f'versus baseline.\n')

    else:
        categorical_variables = [x.lower() for x in config.categorical_or_binary_vars]

        for param, value in baseline_params[0].items():
            if param in categorical_variables:
                baseline_params[0][param] = f'"{value}"'

        baseline_param_query = " & ".join(" == ".join((str(k), str(v))) for k, v in baseline_params[0].items())
        non_baseline_params_query = " | ".join(" != ".join((str(k), str(v))) for k, v in baseline_params[0].items())

    table.columns = columns
    return [baseline_param_query, non_baseline_params_query], table.columns[list(critical_column_locs)]


def structure_data(config, participant_paths):
    excluded_variables = '\n'.join([f'{x}: {y}' for x, y in zip(config.parameter_dict1, config.exclude_data_statistics) if y])
    Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                         '\nRemoving data associated with variables:\n'
                         f"{excluded_variables}")

    combined_results, combined_results_path = Utils.read_combined_results(os.getcwd(), config.averaging_type)

    all_subject_result_jsons = []  # Used for lmm
    averaged_subject_result_jsons = []  # Used for t-tests

    for path in participant_paths:
        all_subject_result_jsons.extend(
            [f"{path}/Summarised_results/{jsn}" for jsn in Utils.find_files(f"{path}/Summarised_results/", "json") if
             'combined_results' not in jsn])

        averaged_subject_result_jsons.extend(
            [f"{path}/Summarised_results/Averaged_results/{jsn}" for jsn in
             Utils.find_files(f"{path}/Summarised_results/Averaged_results/", "json")
             if 'combined_results' not in jsn])

    individual_roi_results = Histogram.load_and_restructure_jsons(config, all_subject_result_jsons, combined_results,
                                                                  data_type='statistics')

    averaged_roi_results = Histogram.load_and_restructure_jsons(config, averaged_subject_result_jsons,
                                                                combined_results,
                                                                data_type='statistics')

    individual_roi_results = Utils.exclude_variables(individual_roi_results, config.exclude_data_statistics)
    averaged_roi_results = Utils.exclude_variables(averaged_roi_results, config.exclude_data_statistics)
    combined_results = Utils.exclude_variables(combined_results, config.exclude_data_statistics)

    return individual_roi_results, averaged_roi_results, combined_results, combined_results_path


def split_dict_by_roi(result_dict):
    """Split combined dataframe into a dict entry for each ROI."""
    ROI_dict = {}
    grouped = result_dict.groupby(result_dict.ROI)
    for ROI in result_dict.ROI.unique():
        ROI_dict[f'{ROI}'] = grouped.get_group(ROI).dropna()
        ROI_dict[f'{ROI}'].columns = [x.replace(' ', '_') for x in ROI_dict[f'{ROI}'].columns]

    return ROI_dict


def statistics(param_queries, critical_params):
    individual_roi_results, averaged_roi_results, raw_data, combined_roi_results, combined_roi_results_path = load_data()
    main_effect_parameters, simple_effect_parameters, excluded_parameters = extract_statistics_options_info()

    # Get list of rois before cleaning up the dataframe
    list_rois = list(combined_roi_results['index'].unique())

    rois_below_max_r2_thresh, individual_roi_results, averaged_roi_results, raw_data, combined_roi_results \
        = clean_results(individual_roi_results, excluded_parameters, raw_data,
                        averaged_roi_results, combined_roi_results,
                        voxel_cutoff=config.minimum_voxels)

    if '' not in param_queries:  # If no baseline parameter combination set, skip this
        calculate_percent_change_versus_baseline(param_queries, combined_roi_results, combined_roi_results_path)

    if config.run_t_tests or config.run_linear_mixed_models:
        if not [parameter for parameter in [*main_effect_parameters.values(), *simple_effect_parameters.values()] if
                parameter] \
                and config.run_t_tests:
            warnings.warn('No main or simple effects chosen in statisticsOptions.csv.'
                          '\nUpdate this file to run t-tests.')

        chosen_rois = Utils.find_chosen_rois(list_rois, func_name="Statistics",
                                             config_region_var=config.regional_stats_rois)

        roi_statistics(critical_params, rois_below_max_r2_thresh,
                       individual_roi_results, averaged_roi_results,
                       chosen_rois, raw_data,
                       main_effect_parameters, simple_effect_parameters)


def load_data():
    participants, _ = Utils.find_participant_dirs(os.getcwd())

    individual_roi_results, averaged_roi_results, combined_results, combined_results_path = structure_data(config, participants)

    individual_roi_results = split_dict_by_roi(individual_roi_results)
    averaged_roi_results = split_dict_by_roi(averaged_roi_results)

    raw_data = copy.deepcopy(individual_roi_results)

    return individual_roi_results, averaged_roi_results, raw_data, combined_results, combined_results_path


def roi_statistics(critical_params, rois_below_max_r2_thresh,
                   individual_roi_results_dict, averaged_roi_results,
                   chosen_rois, raw_data, main_effect_parameters, simple_effect_parameters):
    glm_formula_types = []

    if config.main_effects:
        glm_formula_types.append('main_effects')
    if config.main_and_interaction_effects:
        glm_formula_types.append('main_and_interaction_effects')
    if config.interaction_effects:
        glm_formula_types.append('interaction_effects')

    for glm_formula_type in glm_formula_types:
        standardised_coeffs_df = pd.DataFrame()
        unstandardised_coeffs_df = pd.DataFrame()

        Utils.check_and_make_dir(f"{STATISTICS_PATH}/{glm_formula_type}")
        formula = construct_glm_formula(critical_params, glm_formula_type)

        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f'\n==============================================================================\n'
                             f'\nUsing formula: "{formula}" for "{glm_formula_type.replace("_", " ")}" calculations.\n'
                             f'\n==============================================================================\n')

        adj_marginal_r_square_data = []
        adj_conditional_r_square_data = []
        voxel_data = []

        for ROI in chosen_rois:
            Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                                 f'\n==============================================================================\n'
                                 f'\nCalculating {ROI} statistics ({glm_formula_type.replace("_", " ")}).\n')

            if not individual_roi_results_dict[ROI].empty:
                if config.run_linear_mixed_models:
                    individual_roi_results_dict[ROI].columns = individual_roi_results_dict[ROI].columns.str.lower()

                    adj_marginal_r_square, \
                    adj_conditional_r_square, \
                    converged, \
                    standardised_coeffs, \
                    unstandardised_coeffs, \
                    pvalues = run_glm(individual_roi_results_dict[ROI], ROI=ROI, formula=formula,
                                      glm_formula_type=glm_formula_type)

                    if adj_marginal_r_square is not None \
                            and ROI not in ['Overall', 'No ROI'] \
                            and ROI not in rois_below_max_r2_thresh \
                            and converged:
                        adj_marginal_r_square_data.append(adj_marginal_r_square)
                        adj_conditional_r_square_data.append(adj_conditional_r_square)
                        voxel_data.append(individual_roi_results_dict[ROI].mean(numeric_only=True)['voxel_amount'])

                        # Save list of coefficients to make brain images
                        unstandardised_coeffs = unstandardised_coeffs.loc[pvalues <= config.brain_map_p_thresh]
                        standardised_coeffs = standardised_coeffs.loc[pvalues <= config.brain_map_p_thresh]

                        if not standardised_coeffs.empty:
                            standardised_coeffs['ROI'] = ROI
                            standardised_coeffs_df = pd.concat(
                                [standardised_coeffs_df, standardised_coeffs[['standardised coef', 'ROI']]])

                        if not unstandardised_coeffs.empty:
                            unstandardised_coeffs = pd.DataFrame(unstandardised_coeffs, columns=['unstandardised coef'])
                            unstandardised_coeffs['ROI'] = ROI
                            unstandardised_coeffs_df = pd.concat([unstandardised_coeffs_df, unstandardised_coeffs])

                if config.run_t_tests:
                    averaged_roi_results[ROI].columns = averaged_roi_results[ROI].columns.str.lower()
                    raw_data[ROI].columns = raw_data[ROI].columns.str.lower()

                    run_t_tests(averaged_roi_results[ROI], critical_params, raw_data[ROI],
                                main_effect_parameters, simple_effect_parameters, ROI)

            else:
                Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                                     'All sessions below minimum voxel count.\n'
                                     'Skipping statistics for this ROI.\n')

        if voxel_data:
            r_square_dict = {'adj_marginal_rsquare': adj_marginal_r_square_data,
                             'adj_conditional_rsquare': adj_conditional_r_square_data}

            for rsquare_type in r_square_dict:
                compute_rsquare_regression(voxel_data, r_square_dict[rsquare_type], rsquare_type, glm_formula_type)

            if config.verbose:
                print("\n=============================================================================="
                      "\nCreating standardised and unstandardised coefficient brain maps"
                      "\n==============================================================================\n")

            if not standardised_coeffs_df.empty:
                Coefficient_map.to_braingrid(standardised_coeffs_df, "standardised_coeffs",
                                             standardise_cbar=True, subsubfolder=glm_formula_type)
            elif config.verbose:
                print('Skipping creation of standardised coefficient brain maps. All ROIs below p-value threshold.')

            if not unstandardised_coeffs_df.empty:
                Coefficient_map.to_braingrid(unstandardised_coeffs_df, "unstandardised_coeffs",
                                             standardise_cbar=False, subsubfolder=glm_formula_type)
            elif config.verbose:
                print('Skipping creation of unstandardised coefficient brain maps. All ROIs below p-value threshold.')
        elif config.verbose:
                print('Skipping creation of coefficient brain maps. No suitable ROIs found to plot.')


def compute_rsquare_regression(voxel_data, r_square_data, r_square_type, glm_formula_type):
    if len(voxel_data) == 1:
        if config.verbose:
            print('\nOnly one ROI passed to rsquare regression function, not able to calculate result.')

        return

    # GLM
    voxel_data = sm.add_constant(voxel_data)
    voxel_df = pd.DataFrame(columns=['Intercept', 'Voxels'], data=voxel_data)
    r_square_df = pd.DataFrame(columns=['R_Square'], data=r_square_data)

    model = sm.OLS(r_square_df, voxel_df)
    result = model.fit()

    formula = construct_glm_formula(critical_params=['Voxels'], glm_formula_type=glm_formula_type)
    standardised_coeffs = calculate_glm_standardised_coeffs(pd.concat([voxel_df[['Voxels']], r_square_df], axis=1),
                                                            'R2', formula, glm_formula_type, variation=r_square_type)

    Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                         f'\n==============================================================================\n'
                         f'====================== {r_square_type.replace("_", " ")} results =======================\n'
                         f'Calculating r2 vs mean voxel amount statistics for {glm_formula_type.replace("_", " ")}.\n'
                         f'Note: ROIs with more than {config.max_below_thresh}% of sessions excluded (due to them\n'
                         f'being under the minimum voxel count), or ROIs '
                         f'where the model failed to converge, are excluded from this analysis.\n\n'
                         f'{result.summary()}\n\n\n'
                         f'Standardised coefficients:\n'
                         f'{standardised_coeffs}')

    with open(f"{STATISTICS_PATH}/{glm_formula_type}/{r_square_type}_vs_voxels_GLM.csv", "w") as f:
        f.write(result.summary().as_csv())

    # Scatter plot
    df = pd.concat([voxel_df, r_square_df], axis=1)
    figure = (
            pltn.ggplot(df, pltn.aes(x="Voxels", y="R_Square"))
            + pltn.theme_538()
            + pltn.geom_point()
            + pltn.geom_smooth(method='lm')
            + pltn.labs(x='Mean voxel count', y='$r^2$')
            + pltn.theme(
        panel_grid_minor_x=pltn.themes.element_line(alpha=0),
        panel_grid_major_x=pltn.themes.element_line(alpha=1),
        panel_grid_major_y=pltn.element_line(alpha=0),
        plot_background=pltn.element_rect(fill="white"),
        panel_background=pltn.element_rect(fill="gray", alpha=0.1),
        axis_title_x=pltn.element_text(weight='bold', color='black', size=20),
        axis_title_y=pltn.element_text(weight='bold', color='black', size=20),
        strip_text_x=pltn.element_text(weight='bold', size=10, color='black'),
        strip_text_y=pltn.element_text(weight='bold', size=10, color='black'),
        axis_text_x=pltn.element_text(size=10, color='black'),
        axis_text_y=pltn.element_text(size=10, color='black'),
        dpi=config.plot_dpi
    )
    )

    figure.save(f"{STATISTICS_PATH}/{glm_formula_type}/{r_square_type}_vs_voxels.png",
                height=config.plot_scale,
                width=config.plot_scale * 2,
                verbose=False, limitsize=False)

    figure.save(f"{STATISTICS_PATH}/{glm_formula_type}/{r_square_type}_vs_voxels.svg",
                height=config.plot_scale,
                width=config.plot_scale * 2,
                verbose=False, limitsize=False)


def calculate_percent_change_versus_baseline(param_queries, combined_results, combined_results_path):
    if config.verbose:
        print(f'\nCalculating percentage change statistics.\n')

    combined_results['Percentage change from baseline'] = np.nan
    combined_results['Baseline'] = ""
    combined_results = combined_results.sort_values('index')

    columns = combined_results.columns  # Save old column names before converting to lower case

    # Need to convert to lowercase as the df query is easier to make when lowercase
    combined_results.columns = combined_results.columns.str.lower()

    # Add percentage change statistics to original combined_results df
    combined_results = calculate_percentage_change_per_roi(combined_results, param_queries)

    bootstrap_df = bootstrap_overall_percentage_change(combined_results)
    save_dfs(bootstrap_df, columns, combined_results, combined_results_path)


def run_t_tests(dataset, critical_params, raw_data, main_effect_parameters, simple_effect_parameters, ROI):
    Utils.check_and_make_dir(f"{STATISTICS_PATH}/Overall")

    number_of_simple_effect_params = len(
        [parameter for parameter in simple_effect_parameters if simple_effect_parameters[parameter]])
    number_of_main_effect_params = len(
        [parameter for parameter in main_effect_parameters if main_effect_parameters[parameter]])

    if number_of_main_effect_params > 0:
        calculate_main_effects(ROI, critical_params, dataset, main_effect_parameters, raw_data)

    if number_of_simple_effect_params > 1:
        calculate_simple_effects(ROI, dataset, simple_effect_parameters)

    elif number_of_simple_effect_params == 1:
        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f"\nOnly one variable setup for simple effect t-tests. Skipping calculation.\n")


def calculate_main_effects(ROI, critical_params, dataset, main_effect_parameters, raw_data):
    t_test_results = []
    for param in critical_params:
        t_test_type = config.IV_type[config.parameter_dict1.index(param)]

        for main_effect_contrast in main_effect_parameters[param]:
            data1 = dataset.loc[dataset[param.lower()].astype(str) == main_effect_contrast.split(' v ')[0]].copy()
            data2 = dataset.loc[dataset[param.lower()].astype(str) == main_effect_contrast.split(' v ')[1]].copy()

            data1, data2, current_result = balance_main_effect_data(critical_params, [data1, data2], param, raw_data,
                                                                    t_test_type)

            if not current_result:
                data1 = data1.groupby('subject').mean(numeric_only=True)['voxel_value'].sort_index()
                data2 = data2.groupby('subject').mean(numeric_only=True)['voxel_value'].sort_index()

            if not current_result and t_test_type == 'Within-subjects':
                current_result = list(ttest_rel(a=data1, b=data2))

                if np.isnan(current_result[0]):
                    current_result[0] = 'One participant not sufficient to run statistical test, analysis not ran'

                # Degrees of freedom calculation
                current_result.append(len(data1) - 1)

            elif not current_result and t_test_type == 'Between-subjects':
                current_result = list(ttest_ind(x1=data1, x2=data2, usevar='unequal'))

            # Insert label of parameter combination being tested
            current_result.insert(0, f"({t_test_type}) {param}: {main_effect_contrast}")

            if not np.isnan(current_result[-1]):
                current_result.append(data1.mean() - data2.mean())
                current_result.append(calculate_cohens_d(x1=data1, x2=data2))

            t_test_results.append(current_result)

    if t_test_results:
        t_test_df = pd.DataFrame(data=t_test_results, columns=['Variable contrast', 'tstat', 'pvalue', 'df',
                                                               'Mean difference (x1 - x2)', "Cohen's d"])
        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f"Main effect t-test results\n"
                             f'{t_test_df}\n')

        t_test_df.to_csv(f"{STATISTICS_PATH}/Overall/{ROI}_main_effect_t_tests.csv", index=False)


def calculate_simple_effects(ROI, dataset, simple_effect_parameters):
    t_test_results = []

    for param in simple_effect_parameters:
        if not simple_effect_parameters[param] or len(simple_effect_parameters[param]) == 1:
            continue

        t_test_type = config.IV_type[config.parameter_dict1.index(param)]

        simple_effect_combinations = list(itertools.combinations(simple_effect_parameters[param], 2))

        for simple_effect_contrast in simple_effect_combinations:
            other_params = {other_param.lower(): simple_effect_parameters[other_param] for other_param in
                            simple_effect_parameters if other_param != param}

            keys, values = zip(*other_params.items())
            for combination in itertools.product(*values):
                current_combination = dict(zip(keys, combination))

                contrast_1, contrast_2 = fix_parameter_trailing_zero_issue_if_float(
                    dataset,
                    param,
                    simple_effect_contrast[0],
                    simple_effect_contrast[1])

                data1 = dataset.loc[dataset[param.lower()].astype(str) == contrast_1].copy()
                data2 = dataset.loc[dataset[param.lower()].astype(str) == contrast_2].copy()

                # Control for each of the other parameters
                for other_param in other_params:
                    controlled_param, = fix_parameter_trailing_zero_issue_if_float(
                        dataset,
                        other_param,
                        current_combination[other_param])

                    data1 = data1.loc[data1[other_param.lower()].astype(str) == controlled_param].copy()
                    data2 = data2.loc[data2[other_param.lower()].astype(str) == controlled_param].copy()

                data1, data2, current_result = balance_simple_effect_data([data1, data2], t_test_type)

                if not current_result:
                    data1, data2 = data1['voxel_value'].sort_index(), data2['voxel_value'].sort_index()

                if not current_result and t_test_type == 'Within-subjects':
                    current_result = list(ttest_rel(a=data1, b=data2))

                    if np.isnan(current_result[0]):
                        current_result[0] = 'One participant not sufficient to run statistical test, analysis not ran'

                    # Degrees of freedom calculation
                    current_result.append(len(data1) - 1)

                elif not current_result and t_test_type == 'Between-subjects':
                    current_result = list(ttest_ind(x1=data1, x2=data2, usevar='unequal'))

                # Insert label of parameter combination being tested
                current_result.insert(0, f"({t_test_type}) {param}: {' v '.join(simple_effect_contrast)}")
                # Insert which variables have been controlled
                current_result.insert(0, str(current_combination))

                if not np.isnan(current_result[-1]):
                    current_result.append(data1.mean() - data2.mean())
                    current_result.append(calculate_cohens_d(x1=data1, x2=data2))

                t_test_results.append(current_result)

    if t_test_results:
        t_test_df = pd.DataFrame(data=t_test_results, columns=['Controlled variables', 'Variable contrast',
                                                               'tstat', 'pvalue', 'df',
                                                               'Mean difference (x1 - x2)', "Cohen's d"])

        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f"Simple effect t-test results\n"
                             f'{t_test_df}\n')

        t_test_df.to_csv(f"{STATISTICS_PATH}/Overall/{ROI}_simple_effect_t_tests.csv", index=False)


def fix_parameter_trailing_zero_issue_if_float(dataset, parameter_name, *argv):
    if dataset[parameter_name.lower()].dtype == 'float64':
        parameters = []

        for arg in argv:
            parameters.append(str(float(arg)))

        return parameters
    else:
        return list(argv)


def balance_simple_effect_data(data_subsets, t_test_type):
    if t_test_type == 'Within-subjects':
        # If data is repeated measures data, remove any participants not in both data subsets
        participants_in_each_df = pd.merge(data_subsets[0]['subject'].drop_duplicates(),
                                           data_subsets[1]['subject'].drop_duplicates(),
                                           how='outer', indicator=True)
        participants_to_remove = participants_in_each_df.loc[(participants_in_each_df['_merge'] == 'left_only')
                                                             | (participants_in_each_df['_merge'] == 'right_only')][
            'subject']

        data_subsets[0] = data_subsets[0][~data_subsets[0]['subject'].isin(participants_to_remove)]
        data_subsets[1] = data_subsets[1][~data_subsets[1]['subject'].isin(participants_to_remove)]

    if data_subsets[0].empty or data_subsets[1].empty:
        current_result = ['No participants remaining after balancing data, analysis not ran',
                          np.NaN, np.NaN, np.NaN, np.NaN]

        return data_subsets[0], data_subsets[1], current_result

    else:
        current_result = []

        return data_subsets[0], data_subsets[1], current_result


def balance_main_effect_data(critical_params, data_subsets, param, raw_data, t_test_type):
    # Compare parameters not currently being compared to see if they are the same between both datasets
    # If not, don't run the analysis
    other_params = {other_param.lower(): config.IV_type[config.parameter_dict1.index(other_param)]
                    for other_param in critical_params
                    if other_param != param}
    within_subject_params = [other_param for other_param in other_params if
                             other_params[other_param] == 'Within-subjects']

    for other_param in other_params.keys():
        if set(data_subsets[0][other_param]) != set(data_subsets[1][other_param]):
            current_result = ['Parameters not equal, analysis not ran', np.NaN, np.NaN, np.NaN, np.NaN]

            return data_subsets[0], data_subsets[1], current_result

    if within_subject_params:
        for data in data_subsets:
            for subject in data['subject'].unique():
                # Check each subject to see if they contain all within-subject parameter combinations seen in full
                # dataset, if not then remove the participant
                if 'left_only' in pd.merge(raw_data[within_subject_params].drop_duplicates(),
                                           data.loc[data['subject'] == subject][within_subject_params],
                                           how='outer', indicator=True)['_merge'].values:
                    rows_to_remove = data.loc[(data['subject'] == subject)].index
                    data.drop(rows_to_remove, inplace=True)

    if t_test_type == 'Within-subjects':
        # If data is repeated measures data, remove any participants not in both data subsets
        participants_in_each_df = pd.merge(data_subsets[0]['subject'].drop_duplicates(),
                                           data_subsets[1]['subject'].drop_duplicates(),
                                           how='outer', indicator=True)
        participants_to_remove = participants_in_each_df.loc[(participants_in_each_df['_merge'] == 'left_only')
                                                             | (participants_in_each_df['_merge'] == 'right_only')][
            'subject']

        data_subsets[0] = data_subsets[0][~data_subsets[0]['subject'].isin(participants_to_remove)]
        data_subsets[1] = data_subsets[1][~data_subsets[1]['subject'].isin(participants_to_remove)]

    if data_subsets[0].empty or data_subsets[1].empty:
        current_result = ['No participants remaining after balancing data, analysis not ran',
                          np.NaN, np.NaN, np.NaN, np.NaN]

        return data_subsets[0], data_subsets[1], current_result

    else:
        current_result = []

        return data_subsets[0], data_subsets[1], current_result


def calculate_cohens_d(x1, x2):
    """Calculation of cohens d using pooled standard deviation. Used for both within and between subjects designs as
    with equal sample sizes for each group, the calculation is the same as for calculation average standard deviation."""
    # Calculate the size of samples
    n1, n2 = len(x1), len(x2)

    # calculate the variance of the samples
    s1, s2 = np.var(x1, ddof=1), np.var(x2, ddof=1)

    # calculate the pooled standard deviation
    dof = n1 + n2 - 2
    sigma = np.sqrt(((n1 - 1) * s1 + (n2 - 1) * s2) / dof)

    # calculate the means of the samples
    u1, u2 = np.mean(x1), np.mean(x2)

    # return the effect size
    return (u1 - u2) / sigma


def calculate_r2_measurements(current_df, result, path_to_folder, ROI):
    # Find variance for residuals, random and fixed effects
    var_resid = result.scale
    var_random_effect = float(result.cov_re.iloc[0])
    var_fixed_effect = result.predict(current_df).var()

    # Calculate total variance
    total_var = var_fixed_effect + var_random_effect + var_resid

    # Calculate marginal and conditional r2
    marginal_r2 = var_fixed_effect / total_var
    conditional_r2 = (var_fixed_effect + var_random_effect) / total_var

    adj_marginal_r2 = 1 - ((1 - marginal_r2) * (result.nobs - 1) / (result.nobs - len(result.params)))
    adj_conditional_r2 = 1 - ((1 - conditional_r2) * (result.nobs - 1) / (result.nobs - len(result.params)))

    headers = ['Marginal R2',
               'Conditional R2',
               'Adj. Marginal R2',
               'Adj. Conditional R2']

    r2 = pd.DataFrame(index=headers,
                      data=[marginal_r2,
                            conditional_r2,
                            adj_marginal_r2,
                            adj_conditional_r2])

    Utils.print_and_save(STATISTICS_LOGFILE, config.print_result, f'R\u00b2 values:\n{r2.to_string(header=False)}',
                         '\n')

    with open(f"{path_to_folder}/{ROI}_r2.csv", "w") as f:
        f.write(r2.to_csv(header=False))

    return adj_marginal_r2, adj_conditional_r2


def calculate_information_criteria(formula, current_df, path_to_folder, ROI):
    model = smf.mixedlm(formula=formula, data=current_df, groups=current_df["subject"])

    try:
        result = model.fit(reml=False)

    except numpy.linalg.LinAlgError as exception:
        information_criteria = pd.DataFrame(index=["Error: Singular matrix"],
                                            data=["Information criteria could not be calculated"])
        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f'{exception} error: skipping ROI.')
    else:
        negative_two_log_likelihood = -2 * result.llf
        number_of_parameters = result.df_modelwc + 1
        aic = result.aic
        bic = result.bic

        headers = ['Converged',
                   '-2 Log Likelihood',
                   'Number of parameters',
                   "Akaike's Information Criterion (AIC)",
                   "Schwarz's Bayesian Criterion (BIC)"]

        information_criteria = pd.DataFrame(index=headers,
                                            data=[result.converged,
                                                  negative_two_log_likelihood,
                                                  number_of_parameters,
                                                  aic,
                                                  bic])

    Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                         f"Information criteria:\n{information_criteria.to_string(header=False)}\n\n")

    with open(f"{path_to_folder}/{ROI}_information_criteria.csv", "w") as f:
        f.write(information_criteria.to_csv(header=False))


def run_glm(dataset, ROI, formula, glm_formula_type):
    adj_marginal_r2, adj_conditional_r2, standardised_coeffs, unstandardised_coeffs, pvalues = None, None, None, None, None
    converged = False

    Utils.check_and_make_dir(f"{STATISTICS_PATH}/{glm_formula_type}/{ROI}")

    current_df = dataset.copy()
    model = smf.mixedlm(formula=formula, data=current_df, groups=current_df["subject"])

    try:
        result = model.fit()

    except numpy.linalg.LinAlgError as exception:
        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f'{exception} error: skipping ROI.')
    else:
        path_to_folder = f"{STATISTICS_PATH}/{glm_formula_type}/{ROI}"

        with open(f"{path_to_folder}/{ROI}_LMM.csv", "w") as f:
            f.write(result.summary().as_text())

        standardised_coeffs = calculate_glm_standardised_coeffs(current_df, ROI, formula, glm_formula_type)

        Utils.print_and_save(STATISTICS_LOGFILE, config.print_result,
                             f'{result.summary()}\n'
                             f'Standardised coefficients:\n'
                             f'{standardised_coeffs}\n')

        converged = result.converged
        unstandardised_coeffs = result.fe_params

        # Get p-values and only keep rows that are in the unstandardised coeffs dataframe (excludes group variance)
        pvalues = result.pvalues.loc[result.pvalues.index.isin(unstandardised_coeffs.index)]

        adj_marginal_r2, adj_conditional_r2 = calculate_r2_measurements(current_df, result, path_to_folder, ROI)
        calculate_information_criteria(formula, current_df, path_to_folder, ROI)

    return adj_marginal_r2, adj_conditional_r2, converged, standardised_coeffs, unstandardised_coeffs, pvalues


def calculate_glm_standardised_coeffs(current_df, folder, formula, glm_formula_type, variation=None):
    # Convert values in numeric columns into zscored values, this will be used to find standardised zscores
    zscored_df = current_df.select_dtypes(include=[np.number]).apply(zscore, ddof=1)

    if config.categorical_or_binary_vars and folder != 'R2':
        categorical_columns = current_df[[x.lower() for x in config.categorical_or_binary_vars]]
        zscored_df = pd.concat([categorical_columns, zscored_df], axis=1)

    if folder == 'R2':
        # Least squares regression
        zscored_model = smf.ols(formula=formula, data=zscored_df)
    else:
        # Linear mixed model
        zscored_model = smf.mixedlm(formula=formula, data=zscored_df, groups=current_df["subject"])

    zscored_result = zscored_model.fit()

    if folder == 'R2':
        path = f"{STATISTICS_PATH}/{glm_formula_type}/{variation}_vs_voxels_GLM_standardised_coeffs.csv"
        standardised_coeffs = pd.DataFrame(
            data=pd.concat([zscored_result.params[1:], zscored_result.conf_int()[1:]], axis=1))
    else:
        path = f"{STATISTICS_PATH}/{glm_formula_type}/{folder}/{folder}_GLM_standardised_coeffs.csv"
        standardised_coeffs = pd.DataFrame(
            data=pd.concat([zscored_result.params[1:-1], zscored_result.conf_int()[1:-1]], axis=1))

    standardised_coeffs.columns = ['standardised coef', '[0.025   ', '   0.975]']

    with open(path, "w") as f:
        f.write(standardised_coeffs.to_csv())

    return standardised_coeffs


def construct_glm_formula(critical_params, glm_formula_type):
    if glm_formula_type == 'main_and_interaction_effects':
        symbol = '*'
    elif glm_formula_type == 'main_effects':
        symbol = '+'
    elif glm_formula_type == 'interaction_effects':
        symbol = ':'

    if critical_params[0] == 'Voxels':
        formula = f'R_Square ~ Voxels'
    else:
        formula = f'voxel_value ~ {f" {symbol} ".join(critical_params).lower()}'

        for param in critical_params:
            if param in config.categorical_or_binary_vars:
                formula = formula.replace(param.lower(), f"C({param.lower()})")

    return formula


def save_dfs(bootstrap_df, columns, combined_results, combined_results_path):
    Utils.check_and_make_dir(f"{STATISTICS_PATH}/Overall")

    bootstrap_df = bootstrap_df.sort_values('Parameter_combination')

    bootstrap_df.to_json(f"{STATISTICS_PATH}/Overall/parameter_percentage_change_vs_baseline.json",
                         orient='records', indent=2)

    combined_results.columns = columns  # Convert columns names back to old versions
    combined_results.to_json(f"{combined_results_path.rsplit('.', 1)[0]}_with_pctchange.json", orient='records',
                             indent=2)


def bootstrap_each_param_comb(combined_results, param_combination, bootstrap_info, current_critical_params):
    columns = {'mean':
                   {'mean': None,
                    f'{config.bootstrap_confidence_interval}_CI': None},
               'percentage change from baseline':
                   {'mean': None,
                    f'{config.bootstrap_confidence_interval}_CI': None}
               }

    percentiles = [(100 - config.bootstrap_confidence_interval) / 2,
                   100 - (100 - config.bootstrap_confidence_interval) / 2]

    # Bootstrap CIs for mean and percentage change
    for column in columns.keys():
        current_df = combined_results.loc[combined_results['file_name'] == param_combination][column].copy()

        # Replace infinite updated data with nan
        current_df.replace([np.inf, -np.inf], np.nan, inplace=True)
        # Drop rows with NaN
        current_df.dropna(inplace=True)

        bootstrap_sample = resample(current_df, n_samples=config.bootstrap_samples)

        bs_sample_mean = []
        for sample in range(len(bootstrap_sample)):
            bs_sample_mean.append(np.mean(bootstrap_sample[sample]))

        columns[column]['mean'] = current_df.mean()
        columns[column][f'{config.bootstrap_confidence_interval}_CI'] = np.percentile(bs_sample_mean,
                                                                                      [percentiles[0],
                                                                                       percentiles[1]])

    bootstrap_info.append([param_combination,
                           *current_critical_params.values(),
                           columns['mean']['mean'], columns['mean'][f'{config.bootstrap_confidence_interval}_CI'],
                           columns['percentage change from baseline']['mean'],
                           columns['percentage change from baseline'][f'{config.bootstrap_confidence_interval}_CI']])

    return bootstrap_info


def bootstrap_overall_percentage_change(combined_results):
    bootstrap_info = []
    for param_combination in combined_results['file_name'].unique():
        critical_params_dict = dict.fromkeys(config.parameter_dict1, None)

        for critical_param in config.parameter_dict1:
            # Find critical parameter values
            critical_params_dict[critical_param] = \
                combined_results.loc[combined_results['file_name'] == param_combination][critical_param.lower()].iloc[0]

        bootstrap_info = bootstrap_each_param_comb(combined_results, param_combination,
                                                   bootstrap_info, critical_params_dict)

    bootstrap_df = pd.DataFrame(data=bootstrap_info,
                                columns=['Parameter_combination', *critical_params_dict.keys(),
                                         'Mean',
                                         f'Bootstrapped_mean_{config.bootstrap_confidence_interval}_CI',
                                         'Percentage_change',
                                         f'Bootstrapped_percentage_change_{config.bootstrap_confidence_interval}_CI'])

    return bootstrap_df


def calculate_percentage_change_per_roi(combined_results, param_queries):
    for ROI in combined_results['index'].unique():
        current_df = combined_results.loc[combined_results['index'] == ROI].copy()

        try:
            baseline_mean = current_df.query(param_queries[0])['mean'].values[0]
        except IndexError:
            continue

        combined_results.loc[current_df.query(param_queries[0]).index, 'percentage change from baseline'] = 0
        combined_results.loc[current_df.query(param_queries[0]).index, 'baseline'] = "y"

        results = current_df.query(param_queries[1])['mean'].apply(
            lambda x: ((x - baseline_mean) / baseline_mean) * 100)
        combined_results.loc[current_df.query(param_queries[1]).index, 'percentage change from baseline'] = results

    return combined_results


def resample(dataset, n_samples):
    resampled = [None] * n_samples  # Create list of size n_samples and fill with None
    dataset = np.asarray(dataset)  # Make sure dataset is in correct format

    for i in range(n_samples):
        resampled[i] = random.choices(dataset, k=len(dataset))  # Resample dataset n_sample number of times

    return np.asarray(resampled)



from functools import reduce

import numpy as np
from scipy.stats import stats

from fRAT.utils import Utils

UTILITY_NAME = "Separate noise volumes"
FOLDER_NAME = ["noise_volume", "func_volumes"]


def run(*args, **kwargs):
    SeparateNoiseVolumes(*args, **kwargs)


class SeparateNoiseVolumes:
    def __init__(self, *args, **kwargs):
        self.config = args[0]
        self.file = args[1]
        self.no_ext_file = args[2]
        self.project_root = args[3]
        self.participant_name = args[4]
        self.participant_dir = args[5]
        self.output_folder = args[6]

        self.separate_noise_from_func()

    def separate_noise_from_func(self):
        noise_data = None
        func_data = None

        data, header = Utils.load_brain(self.file)

        mean_of_timepoints = data.reshape(reduce(lambda x, y: x*y, data.shape[0:3]), data.shape[3]).mean(axis=0)
        outlier = np.where(np.abs(stats.zscore(mean_of_timepoints)) > 3)

        if len(outlier[0]) == 1 and (outlier[0][0] in (0, data.shape[3] - 1)):  # todo hook new folder up to isnr calc
            noise_data = data[:, :, :, outlier]
            func_data = np.delete(data, outlier, axis=3)

        elif len(outlier[0]) == 0:
            print(f'        No noise volume found for file: {self.no_ext_file}')

        else:
            raise Exception(f'Incorrect or multiple potential noise volumes found for file: {self.no_ext_file}.')

        if self.config.verbose and outlier[0][0] == 0:
            print(f'            Noise volume found at beginning of timeseries for file: {self.no_ext_file}')

        elif self.config.verbose and outlier[0][0] == data.shape[3] - 1:
            print(f'            Noise volume found at end of timeseries for {self.no_ext_file}')

        Utils.save_brain(noise_data, '_noise_volume', self.no_ext_file,
                         f"{self.participant_dir}/{FOLDER_NAME[0]}", header)

        Utils.save_brain(func_data, '', self.no_ext_file,
                         f"{self.participant_dir}/{FOLDER_NAME[1]}", header)



import filecmp
import time
import warnings


class TestDifferences:
    def __init__(self, directories, verbose_errors):
        self.directories = directories
        self.verbose_errors = verbose_errors

        self.directory_comparison = None
        self.missing_files_left = 0
        self.missing_files_right = 0
        self.status = ''

        self.warnings_setup()
        self.run_file_comparison()

    def warnings_setup(self):
        if '.DS_Store' not in filecmp.DEFAULT_IGNORES:
            filecmp.DEFAULT_IGNORES.append('.DS_Store')

        warnings.formatwarning = self.warning_on_one_line

    def run_file_comparison(self):
        print(f'\nComparing directories {self.directories[0]} & {self.directories[1]}.')
        time.sleep(0.5)  # Sleep so output does not overlap

        directory_comparison = filecmp.dircmp(self.directories[0], self.directories[1])

        self.print_differences(directory_comparison)
        self.print_final_results()

    def print_differences(self, directory_comparison):
        for left_only in directory_comparison.left_only:
            self.missing_files_right += 1

            if self.verbose_errors == 'true':
                warnings.warn(f"Missing file found in {directory_comparison.right}: {left_only}")

        for right_only in directory_comparison.right_only:
            self.missing_files_left += 1

            if self.verbose_errors == 'true':
                warnings.warn(f"Missing file found in {directory_comparison.left}: {right_only}")

        for sub_folder in directory_comparison.subdirs.values():
            self.print_differences(sub_folder)

    def print_final_results(self):
        if self.missing_files_left == 0 and self.missing_files_right == 0:
            print(f'No missing files found.')
        else:
            if self.missing_files_left != 0:
                warnings.warn(f"{self.missing_files_left} missing files found in {self.directories[0]}.")
            else:
                print(f'No missing files found in {self.directories[0]}.')

            if self.missing_files_right != 0:
                warnings.warn(f"{self.missing_files_right} missing files found in {self.directories[1]}.")
            else:
                print(f'No missing files found in {self.directories[1]}.')

        if self.missing_files_left == 0 and self.missing_files_right == 0:
            self.status = 'No errors'
        else:
            self.status = 'Errors found'

    @staticmethod
    def warning_on_one_line(message, category, filename, lineno, file):
        return f'{message}\n'




import dominate
import os
from dominate.tags import *
from glob import glob
from .utils import Utils


def create_index():
    doc = doc_setup("index.html")

    with doc.body:
        navbar(index=True)

        with div(cls='container'):
            with div(id='content'):
                for i in range(5):
                    br()

                h2('Figures')

                br()

                with table():
                    with thead():
                        with tr():
                            pass

                    with tbody():
                        # Find all figure directories that contain png images
                        fig_types = [f for f in glob(f"Figures/*") if glob(f"{f}/**/*.png", recursive=True)]

                        for fig in fig_types:
                            l = tr()
                            l.add(td(a(h4(str_format(os.path.split(fig)[1])), href=f'fRAT_report/{os.path.split(fig)[1]}.html')))
                            image = glob(f"{fig}/**/*.png", recursive=True)[0]
                            l.add(td(a(img(src=f"{image}", width=400), href=f'fRAT_report/{os.path.split(fig)[1]}.html')))

    save_page(doc, 'index.html')


def create_figure_pages(figure_type):
    folders = [f for f in glob(f"{figure_type}/*/", recursive=True)]

    if len(folders):
        figures = []
        for folder in folders:
            figures.append([f for f in glob(f"{folder}*.png", recursive=True)])

    else:
        folders = [figure_type]
        figures = [f for f in glob(f"{folders[0]}/**/*.png", recursive=True)]

    for counter, folder in enumerate(folders):
        doc = doc_setup()
        axis_type = os.path.split(folders[counter][:-1])[1]

        with doc.body:
            navbar()

            with div(cls='container'):
                with div(id='content'):
                    for i in range(5):
                        br()

                    if len(folders) > 1:
                        folder = os.path.split(os.path.split(folder[:-1])[0])[1]
                        h2(f"{folder} ({str_format(axis_type).lower()})")
                    else:
                        folder = os.path.split(folder)[1]
                        h2(f"{str_format(folder)}")

                    br()

                    with table():
                        with tbody():
                            if len(folders) > 1:
                                plot_figs(folder, figures[counter])
                            else:
                                plot_figs(folder, figures)

        if len(folders) > 1:
            save_page(doc, f'{os.path.split(figure_type)[1]}_{axis_type}.html')
        else:
            save_page(doc, f'{os.path.split(figure_type)[1]}.html')

    if len(folders) > 1:
        doc = doc_setup()

        with doc.body:
            navbar()

            with div(cls='container'):
                with div(id='content'):
                    for i in range(5):
                        br()

                    h2(os.path.split(figure_type)[1])

                    br()

                    with table():
                        with tbody():
                            l = tr()

                            for folder in folders:
                                l.add(a(h4(str_format(os.path.split(folder[:-1])[1])),
                                        href=f'{os.path.split(figure_type)[1]}_{os.path.split(folder[:-1])[1]}.html'))

        save_page(doc, f'{os.path.split(figure_type)[1]}.html')


def str_format(string):
    replace_chars = (("axis", "-axis"),
                     ("roi", "ROI"),
                     ("stat", "Stat"),
                     ("same_ylim", ""),
                     ("same_xlim", ""),
                     ('_', ' '))

    try:
        for char in replace_chars:
            string = string.replace(char[0], char[1])
    except AttributeError:
        return string

    return string


def plot_figs(figure_type, figures):
    # Determine how much to trim off the end of the file name (removing .png etc.)
    if figure_type in ('Brain_grids', 'Brain_images'):
        length = 1
    else:
        length = len(figure_type)

    for fig in figures:
        try:
            fig_name = os.path.split(str_format(fig))[1][:-(length+3)]  # Format name for better presentation
        except TypeError:
            fig_name = 'Error: Plotting was stopped before completion.'

        l = tr()
        l.add(td(a(h4(str_format(fig_name)), href=f"../{fig}")))  # Add plot title
        l.add(td(a(img(src=f"../{fig}", width=700), href=f"../{fig}")))  # Show plot


def doc_setup(page=None):
    if page == 'index.html':
        folder = "fRAT_report/"
    else:
        folder = ""

    doc = dominate.document(title='fRAT report')

    with doc.head:
        link(rel='stylesheet', href=f'{folder}bootstrap.css')
        script(type='text/javascript', src=f'{folder}script.js')

    return doc


def save_page(doc, page):
    if page == 'index.html':
        folder = ""
    else:
        folder = "fRAT_report/"

    with open(f"{folder}{page}", 'w') as file:
        file.write(doc.render())


def navbar(index=False):
    with nav(cls="navbar navbar-expand-lg fixed-top navbar-dark bg-primary"):
        with div(cls='container'):
            span('fRAT report', style="font-size:30px", cls="navbar-brand")
            span(id="output", style="font-size:15px", cls="navbar-brand")

            if index:
                with div(cls='collapse navbar-collapse', id='navbarResponsive'):
                    with ul(cls='navbar-nav'):
                        li()  # Prevent flex from forcing folder name to the right hand side on index page

                script("""
                       var loc = window.location.pathname;
                       document.getElementById('output').innerHTML = '/' + loc.split('/').slice(-2, -1);
                       """)
            else:
                with div(cls='collapse navbar-collapse', id='navbarResponsive'):
                    with ul(cls='navbar-nav'):
                        li(a('Back to Figures', cls="nav-link", href=f'../index.html', align='center'), cls='nav-item')

                script("""
                       var loc = window.location.pathname;
                       document.getElementById('output').innerHTML = '/' + loc.split('/').slice(-3, -2);
                       """)


def main(orig_path):
    Utils.check_and_make_dir("fRAT_report")
    orig_path += "/utils"
    Utils.move_file('bootstrap.css', orig_path, 'fRAT_report/', copy=True, rename_copy=False)
    Utils.move_file('script.js', orig_path, 'fRAT_report/', copy=True, rename_copy=False)

    create_index()
    fig_types = [f for f in glob(f"Figures/*") if f != 'Figures/figure_log.toml']

    for fig in fig_types:
        create_figure_pages(fig)



import numpy as np
import pandas as pd

from fRAT.utils import Utils

UTILITY_NAME = "Add Gaussian noise"
FOLDER_NAME = "added_noise"


def run(*args, **kwargs):
    AddNoise(*args, **kwargs)


class AddNoise:
    def __init__(self, *args, **kwargs):
        self.config = args[0]
        self.file = args[1]
        self.no_ext_file = args[2]
        self.project_root = args[3]
        self.participant_name = args[4]
        self.participant_dir = args[5]
        self.output_folder = args[6]

        self.data, self.header = Utils.load_brain(self.file)
        self.add_noise_to_file()

    def add_noise_to_file(self):
        # Get noise level for participant
        noise_value_csv = pd.read_csv(f'{self.project_root}/noiseValues.csv')
        participant_noise_level = float(
            noise_value_csv[noise_value_csv['Participant'] == self.participant_name]['Noise over time'])

        # Save initial data with no added noise
        Utils.save_brain(self.data, ext=f'_noiselevel0', no_ext_file=self.no_ext_file,
                         output_folder=f"{self.participant_dir}/{self.output_folder}", header=self.header)

        # Apply noise multiplier to gaussian noise
        for multiplier in self.config.noise_multipliers:
            gaussian_noise = np.random.default_rng().normal(loc=0.0,
                                                            scale=participant_noise_level * multiplier,
                                                            size=self.data.shape)

            noisy_data = self.data + gaussian_noise
            noisy_data[noisy_data < 0] = 0  # Set values below 0 to 0

            Utils.save_brain(noisy_data, ext=f'_noiselevel{multiplier}', no_ext_file=self.no_ext_file,
                             output_folder=f"{self.participant_dir}/{self.output_folder}", header=self.header)



from fRAT.utils import Utils

import os
import glob
import shutil
import numpy as np
import pandas as pd
from math import sin, cos
from nipype.interfaces import fsl
from nipype.interfaces.base import CommandLineInputSpec, File, TraitedSpec, traits, isdefined
from nipype.interfaces.fsl.base import FSLCommand, FSLCommandInputSpec
from os import path

UTILITY_NAME = "Add motion"
FOLDER_NAME = "added_motion"


def run(*args, **kwargs):
    add_motion_obj = AddMotion(*args, **kwargs)

    if not kwargs['return_val']:
        mean_vals = add_motion_obj.calculate_mean_motion_in_file()

        return mean_vals

    else:
        add_motion_obj.add_motion_to_file()


class ApplyXfm4DInputSpec(FSLCommandInputSpec):
    in_file = File(exists=True, position=0, argstr='%s',
                   mandatory=True, desc="timeseries to motion-correct")
    ref_vol = File(exists=True, position=1, argstr='%s',
                   mandatory=True, desc="volume with final FOV and resolution")
    out_file = File(position=2, argstr='%s',
                    genfile=True, desc="file to write", hash_files=False)
    trans_dir = File(argstr='%s', position=3,
                     desc="folder of transformation matricies")
    four_digit = traits.Bool(
        argstr='-fourdigit', desc="true mat names have four digits not five")
    interpolation = traits.Enum(
        "spline",
        "nn",
        "sinc",
        "trilinear",
        argstr="-interp %s",
        desc="interpolation method for transformation",
    )


class ApplyXfm4DOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="transform applied timeseries")


class ApplyXfm4D(FSLCommand):
    """
    Wraps the applyxfm4D command line tool for applying one 3D transform to every volume in a 4D image OR
    a directory of 3D transforms to a 4D image of the same length.
    """

    _cmd = 'applyxfm4D'
    input_spec = ApplyXfm4DInputSpec
    output_spec = ApplyXfm4DOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_outfilename()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            return self._gen_outfilename()
        return None

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if isdefined(out_file):
            out_file = path.realpath(out_file)
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = self._gen_fname(
                self.inputs.in_file, suffix='_warp4D')
        return path.abspath(out_file)


class AddMotion:
    def __init__(self, *args, **kwargs):
        self.config = args[0]
        self.file = args[1]
        self.no_ext_file = args[2]
        self.project_root = args[3]
        self.participant_name = args[4]
        self.participant_dir = args[5]
        self.output_folder = args[6]

        self.mean_vals = kwargs['return_val']

    def calculate_mean_motion_in_file(self):
        fsl.MCFLIRT(in_file=self.file, out_file=f"{self.participant_dir}/{self.no_ext_file}_mcf.nii.gz",
                    save_plots=True).run()

        df = pd.read_csv(f"{self.participant_dir}/{self.no_ext_file}_mcf.nii.gz.par", header=None,
                         sep="  ", engine='python')
        df = df.loc[~(df == 0).all(axis=1)]

        return list(df.abs().mean(axis=0))

    def add_motion_to_file(self):
        self.data, self.header = Utils.load_brain(self.file)
        self.mean_vals = np.array(self.mean_vals).mean(axis=1)[0]

        timepoints = self.data.shape[-1]

        # Save initial data with no added motion
        Utils.save_brain(self.data, ext=f'_motionlevel0', no_ext_file=self.no_ext_file,
                         output_folder=f"{self.participant_dir}/{self.output_folder}", header=self.header)

        fsl.maths.MeanImage(in_file=self.file, dimension='T',
                            out_file=f"{self.participant_dir}/{self.no_ext_file}_mcf_meaned.nii.gz").run()

        # Apply noise multiplier to gaussian noise
        for multiplier in self.config.motion_multipliers:
            matrix_file_directory = f"{self.participant_dir}/{self.no_ext_file}_MAT_{multiplier}"
            Utils.check_and_make_dir(matrix_file_directory)

            params = []

            for param in self.mean_vals:
                params.append(np.random.default_rng().normal(loc=0.0, scale=param * multiplier, size=timepoints))

            # Transpose matrix so values are grouped per timepoint, not per parameter
            transposed_params = np.array(params).transpose()

            for timepoint, current_params in enumerate(transposed_params):
                transformation_matrix = self.create_transformation_matrix(current_params)
                np.savetxt(f"{matrix_file_directory}/MAT_{str(timepoint).zfill(4)}", transformation_matrix)

            ApplyXfm4D(in_file=self.file, trans_dir=matrix_file_directory,
                       out_file=f"{self.participant_dir}/{self.output_folder}/{self.no_ext_file}_motionlevel{multiplier}.nii.gz",
                       ref_vol=f"{self.participant_dir}/{self.no_ext_file}_mcf_meaned.nii.gz",
                       four_digit=True,
                       interpolation="trilinear",
                       terminal_output='none').run()

        self.clean_up_files()

    @staticmethod
    def create_transformation_matrix(params):
        R_x = np.array([[1, 0, 0, 0],
                        [0, cos(params[0]), sin(params[0]), 0],
                        [0, -sin(params[0]), cos(params[0]), 0],
                        [0, 0, 0, 1]
                        ])

        R_y = np.array([[cos(params[1]), 0, -sin(params[1]), 0],
                        [0, 1, 0, 0],
                        [sin(params[1]), 0, cos(params[1]), 0],
                        [0, 0, 0, 1]
                        ])

        R_z = np.array([[cos(params[2]), sin(params[2]), 0, 0],
                        [-sin(params[2]), cos(params[2]), 0, 0],
                        [0, 0, 1, 0],
                        [0, 0, 0, 1]
                        ])

        T = np.array([[1, 0, 0, params[3]],
                      [0, 1, 0, params[4]],
                      [0, 0, 1, params[5]],
                      [0, 0, 0, 1]
                      ])

        return T @ R_x @ R_y @ R_z

    def clean_up_files(self):
        for mcf in glob.glob(f'{self.participant_dir}/{self.no_ext_file}_mcf*'):
            os.remove(mcf)

        for mat_dir in glob.glob(f'{self.participant_dir}/{self.no_ext_file}_MAT*'):
            shutil.rmtree(mat_dir)



import copy
import itertools
import logging
import os
import pathlib
import re
import shutil
import warnings
from glob import glob
from random import choice

import bootstrapped.bootstrap as bs
import bootstrapped.stats_functions as bs_stats
import nibabel as nib
import numpy as np
import pandas as pd
import simplejson as json
import xmltodict
from nipype.interfaces import fsl
from nipype.interfaces.fsl import maths
from scipy.stats import norm

from .utils import Utils

config = None


class Environment_Setup:
    file_list = []
    save_location = None
    roi_stat_list = ["Voxel number", "Mean", "Standard Deviation", "Confidence Interval", "Min", "Max"]
    conf_level_list = [('80', 1.28),
                       ('85', 1.44),
                       ('90', 1.64),
                       ('95', 1.96),
                       ('98', 2.33),
                       ('99', 2.58)]
    atlas_label_list = [('Cerebellum/Cerebellum-MNIflirt-maxprob-thr0-1mm.nii.gz', 'Cerebellum_MNIflirt.xml'),
                        ('Cerebellum/Cerebellum-MNIfnirt-maxprob-thr0-1mm.nii.gz', 'Cerebellum_MNIfnirt.xml'),
                        ('HarvardOxford/HarvardOxford-cort-maxprob-thr0-1mm.nii.gz', 'HarvardOxford-Cortical.xml'),
                        ('HarvardOxford/HarvardOxford-sub-maxprob-thr0-1mm.nii.gz', 'HarvardOxford-Subcortical.xml'),
                        ('JHU/JHU-ICBM-labels-1mm.nii.gz', 'JHU-labels.xml'),
                        ('JHU/JHU-ICBM-tracts-maxprob-thr0-1mm.nii.gz', 'JHU-tracts.xml'),
                        ('Juelich/Juelich-maxprob-thr0-1mm.nii.gz', 'Juelich.xml'),
                        ('MNI/MNI-maxprob-thr0-1mm.nii.gz', 'MNI.xml'),
                        ('SMATT/SMATT-labels-1mm.nii.gz', 'SMATT.xml'),
                        ('STN/STN-maxprob-thr0-0.5mm.nii.gz', 'STN.xml'),
                        ('Striatum/striatum-structural-1mm.nii.gz', 'Striatum-Structural.xml'),
                        ('Talairach/Talairach-labels-1mm.nii.gz', 'Talairach.xml'),
                        ('Thalamus/Thalamus-maxprob-thr0-1mm.nii.gz', 'Thalamus.xml')]
    base_directory = None
    fsl_path = None
    atlas_path = None
    atlas_label_path = None
    atlas_name = None
    label_array = []

    @classmethod
    def setup_analysis(cls, cfg, config_path, config_filename, pool):
        """Set up environment and find files before running analysis."""
        global config
        config = cfg

        if config.verbose:
            print('\n--- Environment Setup ---')

        cls.setup_environment(config_path, config_filename)

        if config.run_plotting:
            verify_param_values()

        cls.setup_save_location()

        # Extract labels from selected FSL atlas
        cls.atlas_path = f'{cls.fsl_path}/data/atlases/{cls.atlas_label_list[int(config.atlas_number)][0]}'
        cls.atlas_label_path = f'{cls.fsl_path}/data/atlases/{cls.atlas_label_list[int(config.atlas_number)][1]}'

        cls.roi_label_list()

        participant_list = Participant.setup_class(cls.base_directory, cls.save_location, pool)
        matched_brain_list = MatchedBrain.setup_class(participant_list)

        return participant_list, matched_brain_list

    @classmethod
    def setup_save_location(cls):
        cls.atlas_name = os.path.splitext(cls.atlas_label_list[int(config.atlas_number)][1])[0]

        if config.output_folder == 'DEFAULT':
            cls.save_location = f'{cls.atlas_name}_ROI_report/'
        else:
            cls.save_location = f'{config.output_folder}/'

        if config.verbose:
            print(f'Saving output in directory: {cls.save_location}\n'
                  f'Using the {cls.atlas_name} atlas.\n'
                  f'Using parameter file: {config.parameter_file}.\n')

        # Make folder to save ROI_report if not already created
        Utils.check_and_make_dir(f"{cls.base_directory}/{cls.save_location}", delete_old=True)

        # Move config file to analysis folder
        Utils.move_file('analysis_log.toml', cls.base_directory, cls.save_location)

        if config.stat_map_folder == '':
            print('Statistical map folder to use in subject\'s statmap folders not specified.'
                  '\nIf only one folder is found in the statmap directory, this folder will be selected. '
                  '\nMake sure this folder contains the same kind of statistical map for each participant.'
                  '\nIn future consider filling in the statistical map folder field '
                  'as this information will then be added into the analysis_log.toml file for future reference.\n')
        elif config.verbose:
            print(f'Searching for statmaps in directory: statmaps/{config.stat_map_folder}/\n')

    @classmethod
    def setup_environment(cls, config_path, config_filename):
        try:
            cls.fsl_path = os.environ['FSLDIR']
        except (KeyError, OSError) as e:
            raise Exception('FSL environment variable not set.')

        if config.brain_file_loc in ("", " "):
            print('Select the directory of the raw MRI/fMRI brains.')
            cls.base_directory = Utils.file_browser(title='Select the directory of the raw MRI/fMRI brains')

        else:
            cls.base_directory = config.brain_file_loc

            if config.verbose:
                print(f'Finding subject directories in directory: {config.brain_file_loc}\n')

        # Save copy of analysis_log.toml to retain settings. It is saved here as after changing directory it will be
        # harder to find
        Utils.save_config(cls.base_directory, config_path, config_filename,
                          new_config_name='analysis_log',
                          relevant_sections=['General', 'Analysis', 'Parsing'])

        try:
            os.chdir(cls.base_directory)

        except FileNotFoundError:
            raise FileNotFoundError('Chosen base directory of subjects is not a valid directory path.')

    @classmethod
    def roi_label_list(cls):
        """Extract labels from specified FSL atlas XML file."""
        with open(cls.atlas_label_path) as fd:
            atlas_label_dict = xmltodict.parse(fd.read())

        cls.label_array = []
        cls.label_array.append('No ROI')

        for roiLabelLine in atlas_label_dict['atlas']['data']['label']:
            cls.label_array.append(roiLabelLine['#text'])

        cls.label_array.append('Overall')


class Participant:
    def __init__(self, participant, base_directory, save_location):
        self.participant_name = participant
        self.participant_path = f"{base_directory}/{participant}"
        self.save_location = f"{base_directory}/{save_location}{participant}/"
        self.anat_cost = 0
        self.brains = None
        self.file_list = []
        self.anat_head = None
        self.anat_brain = None
        self.anat_to_mni_mat = None
        self.anat_brain_no_ext = None
        self.grey_matter_segmentation = None
        self.white_matter_segmentation = None
        self.all_files_ignored = False
        self.statmap_folder = ''

        self.find_statmap_folder()

    def find_statmap_folder(self):
        if config.stat_map_folder:
            self.statmap_folder = config.stat_map_folder

        else:
            statmap_folders = glob(os.path.join(f"{self.participant_path}/statmaps", '*'))

            if len(statmap_folders) > 1:
                raise FileExistsError(f"No statistical map folder specified, however multiple statistical map folders "
                                      f"found in {self.participant_name}'s statmap folder.")
            elif len(statmap_folders) == 0:
                raise FileNotFoundError(f"No statistical maps found in {self.participant_name}'s statmap folder.")

            else:
                self.statmap_folder = os.path.split(statmap_folders[0])[1]

                if config.verbose:
                    print(
                        f'Searching for {self.participant_name} statmaps in directory: statmaps/{self.statmap_folder}/\n')

    def setup_participant(self, environment_globals, cfg):
        global config
        config = cfg

        self.find_fmri_files()  # Find fMRI files to save in self.brains

        # Make folder to save ROI_report if not already created
        Utils.check_and_make_dir(self.save_location)

        self.anat_setup(environment_globals['fsl_path'])

        # Initialise brain class for each file found
        for counter, brain in enumerate(self.brains):
            self.brains[counter] = Brain(f"{self.participant_path}/{config.input_folder_name}/{brain}",
                                         self.participant_path,
                                         self.participant_name,
                                         self.save_location,
                                         self.anat_head,
                                         self.anat_brain,
                                         self.anat_to_mni_mat,
                                         self.grey_matter_segmentation,
                                         self.white_matter_segmentation,
                                         self.statmap_folder,
                                         environment_globals)

        return self

    @staticmethod
    def setup_class(base_directory, save_location, pool):
        _, participants = Utils.find_participant_dirs(os.getcwd())

        participant_list = set()
        for participant in participants:
            # Initialise participants
            participant_list.add(Participant(participant, base_directory, save_location))

        # Set arguments to pass to run_analysis function
        iterable = zip(participant_list, itertools.repeat("setup_participant"),
                       itertools.repeat(Environment_Setup.__dict__), itertools.repeat(config))

        if config.verbose:
            print(f'\n--- Anatomical file alignment ---')

        # Setup each participant
        if config.multicore_processing:
            participant_list = set(pool.starmap(Utils.instance_method_handler, iterable))

        else:
            participant_list = set(itertools.starmap(Utils.instance_method_handler, iterable))

        return participant_list

    @classmethod
    def find_participant_dirs(cls):
        # Searches for folders that start with p
        participant_dirs = [direc for direc in glob("*") if re.search("^sub-[0-9]+", direc)]

        if len(participant_dirs) == 0:
            raise FileNotFoundError('Participant directories not found.')
        elif config.verbose:
            print(f'Found {len(participant_dirs)} participant folders.')

        return participant_dirs

    def run_analysis(self, pool):
        if config.verbose:
            print(f'\nAnalysing fMRI files for participant: {self.participant_name}\n')

        orig_brain_num = len(self.brains)
        self.brains = [brain for brain in self.brains if
                       brain.parameters]  # Remove any brains that are set to be ignored

        if config.verbose and orig_brain_num != len(self.brains):
            print(f'Ignoring {orig_brain_num - len(self.brains)}/{orig_brain_num} files.\n')

        brain_list = []
        if self.brains:  # self.brains will be empty if all files for this participant have been set to ignore

            # Set arguments to pass to run_analysis function
            iterable = zip(self.brains, itertools.repeat("run_analysis"), range(len(self.brains)),
                           itertools.repeat(len(self.brains)), itertools.repeat(config))

            if config.multicore_processing:
                brain_list = pool.starmap(Utils.instance_method_handler, iterable)
            else:
                brain_list = list(itertools.starmap(Utils.instance_method_handler, iterable))

            construct_combined_results(self.save_location, analysis_type='participant')

            self.anat_file_cleanup()

        else:
            self.all_files_ignored = True
            shutil.rmtree(self.save_location)

        return brain_list

    def find_fmri_files(self):
        # Find all nifti and analyze files
        self.brains = Utils.find_files(f"{self.participant_path}/{config.input_folder_name}", "hdr", "nii.gz", "nii")

        if len(self.brains) == 0:
            raise NameError("No files found.")

    @staticmethod
    def anat_file_check(files, filetype):
        if len(files) > 1:
            raise FileExistsError(f'More than one {filetype} images found in the anat folder.')
        elif len(files) == 0:
            raise FileExistsError(f'No {filetype} images found in the anat folder.')
        else:
            return files[0]

    def anat_setup(self, fsl_path):
        if config.verbose:
            print(f'Converting anatomical file to MNI space for participant: {self.participant_name}')

        anat = Utils.find_files(f'{self.participant_path}/anat/', "hdr", "nii.gz", "nii")

        if config.anat_align_cost_function == 'BBR' and len(anat) < 2:
            raise FileExistsError('When using the BBR cost function for the fMRI to structural registration, the anat '
                                  'folder should contain both the whole head and brain extracted images labelled in '
                                  'the filename with head and brain respectively e.g. "MPRAGE_head", "MPRAGE_brain".')
        elif len(anat) > 2:
            raise FileExistsError(
                'The maximum number of files NIFTI in the anat folder should be two: a whole head and a brain '
                'extracted image, where the brain extracted image has "brain" in the filename e.g. '
                '"MPRAGE_brain". Remove any json or mask files from this folder.'
                '\nNOTE: Wholehead image is only required when '
                'aligning fMRI volume to anatomical volume with the BBR cost function.')

        if len(anat) == 1:
            brain = anat[0]
        else:
            brain = [file for file in anat if 'brain' in file]
            brain = self.anat_file_check(brain, 'brain')

        if config.anat_align_cost_function == 'BBR':
            head = [file for file in anat if 'brain' not in file]
            head = self.anat_file_check(head, 'whole head')

            self.anat_head = f'{self.participant_path}/anat/{head}'

        self.anat_brain = f'{self.participant_path}/anat/{brain}'
        self.anat_brain_no_ext = brain.rsplit(".")[0]
        self.anat_to_mni_mat = fsl_functions(self, self.save_location, self.anat_brain_no_ext,
                                             'FLIRT', self.anat_brain, 'to_mni_from_',
                                             f'{fsl_path}/data/standard/MNI152_T1_1mm_brain.nii.gz')

        if config.grey_matter_segment or config.anat_align_cost_function == 'BBR':
            try:
                self.find_fslfast_files()

            except IndexError:
                if config.run_fsl_fast == 'Run if files not found':
                    if config.verbose:
                        phrases = ['get a cup of coffee',
                                   'respond to those reviewer comments',
                                   'have a good clear out of your inbox',
                                   'read some XKCD comics',
                                   'contemplate the meaning of existence',
                                   'consider writing a short poem',
                                   'listen to Echoes by Pink Floyd',
                                   'mark one of those essays you\'ve been putting off',
                                   'paint the next abstract masterpiece',
                                   'listen to Michigan by Sufjan Stevens',
                                   'get up and have a stretch',
                                   'follow @elliohow on twitter',
                                   'finally learn the Brodmann areas',
                                   'read the Tidy Data paper by Hadley Wickham']

                        print(f'{self.participant_name} missing FSL FAST files. '
                              f'Running FSL FAST (maybe {choice(phrases)}, this may take a while).')

                    # Need to change directory here to get around error caused by nipype trying to find fslfast output
                    # in current working directory
                    orig_direc = os.getcwd()
                    Utils.check_and_make_dir(f'{self.participant_path}/fslfast/')
                    os.chdir(f'{self.participant_path}/fslfast/')

                    fsl_functions(self, f'{self.participant_path}/fslfast/{self.anat_brain_no_ext}',
                                  '', 'FAST', self.anat_brain, '')

                    os.chdir(orig_direc)

                    self.find_fslfast_files()

                else:
                    raise FileNotFoundError(f'fslfast directory for {self.participant_name} does not contain all files'
                                            f'output by FAST.')

    def find_fslfast_files(self):
        self.grey_matter_segmentation = glob(f"{self.participant_path}/fslfast/*_pve_1*")[0]
        self.white_matter_segmentation = glob(f"{self.participant_path}/fslfast/*_pve_2*")[0]

    def calculate_anat_flirt_cost_function(self):
        fslfunc = fsl.FLIRT(in_file=self.anat_brain,
                            schedule=f'{Environment_Setup.fsl_path}/etc/flirtsch/measurecost1.sch',
                            terminal_output='allatonce', dof=config.dof)

        # Calculate MNI cost function value
        mni_cost = run_flirt_cost_function(fslfunc,
                                           f'{Environment_Setup.fsl_path}/data/standard/MNI152_T1_1mm_brain.nii.gz',
                                           f'{self.save_location}Intermediate_files/to_mni_from_{self.anat_brain_no_ext}.mat',
                                           f'{self.save_location}Intermediate_files/{self.anat_brain_no_ext}_mni_redundant.nii.gz',
                                           f'{self.save_location}Intermediate_files/{self.anat_brain_no_ext}_mni_redundant.mat',
                                           config)

        d = {'Participant': [self.participant_name],
             'File': [self.anat_brain_no_ext],
             '(FLIRT to MNI) Cost function value': [mni_cost]}

        df = pd.DataFrame(data=d)

        return df

    def anat_file_cleanup(self):
        """Clean up unnecessary output from either instance of class, or class itself."""
        if config.file_cleanup == 'delete':
            for file in self.file_list:
                os.remove(file)

        elif config.file_cleanup == 'move':
            for file in self.file_list:
                file = file.replace(self.save_location, "")  # Remove folder from start of file name
                Utils.move_file(file, self.save_location, f"{self.save_location}Intermediate_files")

        self.file_list = []


class Brain:
    def __init__(self, brain, participant_folder, participant_name, save_location,
                 anat_head, anat_brain, anat_to_mni_mat, grey_matter_segmentation, white_matter_segmentation,
                 statmap_folder,
                 environment_globals):
        self.brain = brain
        self.save_location = save_location
        self.anat_head = anat_head
        self.anat_brain = anat_brain
        self.anat_to_mni_mat = anat_to_mni_mat
        self.grey_matter_segmentation = grey_matter_segmentation
        self.white_matter_segmentation = white_matter_segmentation
        self.no_ext_brain = Utils.strip_ext(self.brain.split('/')[-1])
        self.stat_brain = f"{participant_folder}/statmaps/{statmap_folder}/{self.no_ext_brain}{config.stat_map_suffix}"
        self.mni_brain = f"{self.save_location}mni_to_{self.no_ext_brain}.nii.gz"
        self.roi_results = None
        self.raw_results_loc = None
        self.roi_stat_list = ""
        self.file_list = []
        self.participant_name = participant_name
        self.parameters = []
        self.noise_threshold = None
        self.lower_gaussian_threshold = None
        self.upper_gaussian_threshold = None
        self.session_number = 0

        # Copying class attributes here is a workaround for dill,
        # which can't access modified class attributes for imported modules.
        self._brain_directory = environment_globals['base_directory']
        self._fsl_path = environment_globals['fsl_path']
        self._atlas_label_path = environment_globals['atlas_label_path']
        self._atlas_name = environment_globals['atlas_name']
        self._labelArray = environment_globals['label_array']
        self.atlas_path = environment_globals['atlas_path']

    def fmri_get_additional_info(self, brain_number_current, brain_number_total, config):
        if config.verbose:
            print(f'Calculating cost function and mean displacement values for volume '
                  f'{brain_number_current + 1}/{brain_number_total}: {self.no_ext_brain}')

        fslfunc = fsl.FLIRT(
            in_file=f'{self.save_location}Intermediate_files/{self.no_ext_brain}/bet_{self.no_ext_brain}.nii.gz',
            schedule=f'{self._fsl_path}/etc/flirtsch/measurecost1.sch',
            terminal_output='allatonce', dof=config.dof)

        anat_cost, mni_cost = None, None

        # Calculate anatomical cost function value
        wmseg = None
        if config.anat_align_cost_function == 'BBR':
            wmseg = f'{self.save_location}Intermediate_files/{self.no_ext_brain}/to_anat_from_{self.no_ext_brain}_fast_wmseg.nii.gz'

        anat_cost = run_flirt_cost_function(fslfunc,
                                            self.anat_brain,
                                            f'{self.save_location}Intermediate_files/{self.no_ext_brain}/to_anat_from_{self.no_ext_brain}.mat',
                                            f'{self.save_location}Intermediate_files/{self.no_ext_brain}_anat_redundant.nii.gz',
                                            f'{self.save_location}Intermediate_files/{self.no_ext_brain}_anat_redundant.mat',
                                            config, wmseg=wmseg)

        # Find absolute and relative mean displacement files
        suffixes = ['_abs_mean.rms', '_rel_mean.rms']
        displacement_vals = []
        for suffix in suffixes:
            with open(f"{self.save_location}Intermediate_files/motion_correction_files/mcf_"
                      f"{self.no_ext_brain}{suffix}", 'r') as file:
                displacement_vals.append(float(file.read().replace('\n', '')))

        d = {'Participant': [self.participant_name],
             'File': [self.no_ext_brain],
             '(FLIRT to MNI) Cost function value': mni_cost,
             '(MCFLIRT) Mean Absolute displacement': displacement_vals[0],
             '(MCFLIRT) Mean Relative displacement': displacement_vals[1]}

        if anat_cost:
            d['(FLIRT to anatomical) Cost function value'] = anat_cost
        if self.noise_threshold:
            d['Noise Threshold'] = self.noise_threshold
        if self.lower_gaussian_threshold:
            d['Lower Gaussian Outlier Threshold'] = self.lower_gaussian_threshold
        if self.upper_gaussian_threshold:
            d['Upper Gaussian Outlier Threshold'] = self.upper_gaussian_threshold

        df = pd.DataFrame(data=d)

        return df

    def run_analysis(self, brain_number_current, brain_number_total, cfg):
        global config
        config = cfg

        logging.getLogger('nipype.interface').setLevel(0)  # Suppress nipype interface terminal output

        if config.verbose:
            print(f'Analysing fMRI volume {brain_number_current + 1}/{brain_number_total}: {self.no_ext_brain}')

        GM_bool = self.roi_flirt_transform(brain_number_current,
                                           brain_number_total)  # Convert MNI brain to native space

        self.roi_stats(GM_bool)  # Calculate and save statistics

        self.file_cleanup()  # Clean up files

        return self

    def file_cleanup(self):
        """Clean up unnecessary output from either instance of class, or class itself."""
        if config.file_cleanup == 'delete':
            for file in self.file_list:
                os.remove(file)

        elif config.file_cleanup == 'move':
            Utils.check_and_make_dir(f"{self.save_location}Intermediate_files")
            Utils.check_and_make_dir(f"{self.save_location}/Intermediate_files/{self.no_ext_brain}")

            for file in self.file_list:
                file = file.replace(self.save_location, "")  # Remove folder from start of file name
                Utils.move_file(file, self.save_location,
                                f"{self.save_location}/Intermediate_files/{self.no_ext_brain}")

        self.file_list = []

        redundant_files = glob(f'{self.save_location}/white_matter_thresholded_{self.no_ext_brain}.nii.gz') \
                          + glob(f'{self.save_location}/{self.no_ext_brain}_init.mat')

        for file in redundant_files:
            os.remove(file)

    def roi_stats_setup(self):
        # Load original brain (with statistical map)
        stat_brain = nib.load(self.stat_brain)
        # Load atlas brain (which has been converted into native space)
        mni_brain = nib.load(self.mni_brain)

        stat_brain = stat_brain.get_fdata()
        mni_brain = mni_brain.get_fdata()

        if mni_brain.shape != stat_brain.shape:
            raise Exception('The matrix dimensions of the standard space and the statistical map brain do not '
                            'match.')

        # Find the number of unique ROIs in the atlas
        roiList = list(range(0, len(self._labelArray) - 1))
        roiNum = np.size(roiList)

        idxBrain = stat_brain.flatten()
        idxMNI = mni_brain.flatten()

        # Create arrays to store the values before and after statistics
        roi_temp_store = np.full([roiNum, idxMNI.shape[0]], np.nan)
        roi_results = np.full([8, roiNum + 1], np.nan)
        roi_results[7, 0:-1] = 0  # Change excluded voxels measure from NaN to 0

        return roi_temp_store, roi_results, idxMNI, idxBrain

    @staticmethod
    def compile_voxel_values(roi_temp_store, roi_results, idxMNI, idxBrain, GM_bool):
        if config.grey_matter_segment:
            non_segmented_volume = roi_temp_store.copy()
        else:
            non_segmented_volume = None

        for counter, roi in enumerate(idxMNI):
            if config.grey_matter_segment:
                non_segmented_volume[int(roi), counter] = idxBrain[counter]

            if not config.grey_matter_segment or GM_bool[counter] == 1:
                roi_temp_store[int(roi), counter] = idxBrain[counter]

            else:
                roi_temp_store[0, counter] = idxBrain[counter]  # Assign to No ROI if voxel is excluded

                if int(roi) != 0:  # If ROI is not 'No ROI' add to excluded voxels list
                    roi_results[7, int(roi)] += 1

        return roi_temp_store, roi_results, non_segmented_volume

    def segmentation_to_fmri(self, anat_aligned_mat, fMRI_volume, brain_number_current, brain_number_total):
        if config.verbose:
            print(f'Aligning fslfast segmentation to fMRI volume {brain_number_current + 1}/{brain_number_total}.')

        prefix = 'fslfast_to_'
        interp = 'trilinear'

        # Save inverse of fMRI to anat
        inverse_mat = fsl_functions(self, self.save_location, self.no_ext_brain, 'ConvertXFM', anat_aligned_mat,
                                    'inverse_anat_to_')

        # Apply inverse of matrix to chosen segmentation to convert it into native space
        segmentation_to_fmri = fsl_functions(self, self.save_location, self.no_ext_brain, 'ApplyXFM',
                                             self.grey_matter_segmentation, prefix, inverse_mat, fMRI_volume, interp)

        return segmentation_to_fmri

    @staticmethod
    def find_gm_from_segment(native_space_segment):
        segment_brain = nib.load(native_space_segment)
        segment_brain = segment_brain.get_fdata().flatten()

        # If voxel has a value below the threshold then set to 1
        grey_matter_bool = (segment_brain >= config.fslfast_min_prob).astype(int)

        return grey_matter_bool

    def roi_flirt_transform(self, brain_number_current, brain_number_total):
        """Function which uses NiPype to transform the chosen atlas into native space."""
        pack_vars = [self, self.save_location, self.no_ext_brain]

        # Motion correction
        fMRI_volume = fsl_functions(*pack_vars, 'MCFLIRT', self.brain, "mcf_")

        # Turn 4D scan into 3D
        fMRI_volume = fsl_functions(*pack_vars, 'MeanImage', fMRI_volume, "mean_")

        # Brain extraction
        fMRI_volume = fsl_functions(*pack_vars, 'BET', fMRI_volume, "bet_")

        if config.verbose:
            print(f'Aligning fMRI volume {brain_number_current + 1}/{brain_number_total} to anatomical volume '
                  f'using {config.anat_align_cost_function} cost function.')

        # Align to anatomical
        if config.anat_align_cost_function == 'BBR':
            white_matter_thresholded = fsl_functions(*pack_vars,
                                                     'maths.Threshold',
                                                     self.white_matter_segmentation,
                                                     "white_matter_thresholded_")

            white_matter_binarised = fsl_functions(*pack_vars,
                                                   'maths.UnaryMaths',
                                                   white_matter_thresholded,
                                                   "", 'binarise_wmseg')

            fmri_to_anat_mat = fsl_functions(*pack_vars, 'EpiReg',
                                             fMRI_volume, "to_anat_from_",
                                             self.anat_brain, self.anat_head,
                                             white_matter_binarised)

        elif config.anat_align_cost_function == 'Correlation Ratio':
            fmri_to_anat_mat = fsl_functions(*pack_vars, 'FLIRT', fMRI_volume, "to_anat_from_", self.anat_brain)

        else:
            raise ValueError('fMRI to anatomical registration cost function type not valid.')

        # Combine fMRI-anat and anat-mni matrices
        mat = fsl_functions(*pack_vars, 'ConvertXFM', fmri_to_anat_mat, 'combined_mat_', 'concat_xfm')

        # Get inverse of matrix
        inverse_mat = fsl_functions(*pack_vars, 'ConvertXFM', mat, 'inverse_combined_mat_')

        # Apply inverse of matrix to chosen atlas to convert it into native space
        fsl_functions(*pack_vars, 'ApplyXFM', self.atlas_path,
                      'mni_to_', inverse_mat, fMRI_volume,
                      'nearestneighbour')

        if config.grey_matter_segment:
            # Convert segmentation to fMRI native space
            segmentation_to_fmri = self.segmentation_to_fmri(fmri_to_anat_mat, fMRI_volume,
                                                             brain_number_current, brain_number_total)

            grey_matter_bool = self.find_gm_from_segment(segmentation_to_fmri)

            return grey_matter_bool

    def roi_stats(self, GM_bool):
        """Function which uses the output from the roi_flirt_transform function to collate the statistical information
        per ROI."""

        # Load brains and pre-initialise arrays
        roi_temp_store, roi_results, idxMNI, idxBrain = self.roi_stats_setup()

        # Combine information from fMRI and MNI brains (both in native space) to assign an ROI to each voxel
        roi_temp_store, roi_results, non_segmented_volume = self.compile_voxel_values(roi_temp_store,
                                                                                      roi_results,
                                                                                      idxMNI,
                                                                                      idxBrain,
                                                                                      GM_bool)

        warnings.filterwarnings('ignore')  # Ignore warnings that indicate an ROI has only nan values

        if config.grey_matter_segment:
            volumes = {'noROI': non_segmented_volume, 'nonGreyMatter': roi_temp_store}

        else:
            volumes = {'noROI': roi_temp_store}

        roi_results, roi_temp_store = self.noise_correct_data_and_create_excluded_voxel_maps(roi_results,
                                                                                             roi_temp_store, volumes)

        # Compile ROI statistics
        roi_results = self.compile_roi_stats(roi_temp_store, roi_results)

        warnings.filterwarnings('default')  # Reactivate warnings

        # if config.bootstrap: # TODO: uncomment this when bootstrapping reimplemented
        #     roi_results = roi_stats_bootstrap(roi_temp_store, roi_results, roiNum, brain_number_current,
        #                                           brain_number_total)  # Bootstrapping

        self.raw_results_loc = reformat_and_save_raw_data(roi_temp_store, self._labelArray,
                                                          self.save_location, self.parameters,
                                                          session_number=self.session_number)  # Save results

        self.save_roi_results(roi_results)

        # Retain variable for atlas_scale function
        self.roi_results = roi_results

    def save_roi_results(self, roi_results):
        headers = ['Voxels', 'Mean', 'Std_dev',
                   f'Conf_Int_{Environment_Setup.conf_level_list[int(config.conf_level_number)][0]}',
                   'Median', 'Min', 'Max', 'Excluded_Voxels']

        # Save results as dataframe
        results = pd.DataFrame(data=roi_results,
                               index=headers,
                               columns=self._labelArray)

        summary_results_path = f"{self.save_location}Summarised_results/"
        Utils.check_and_make_dir(summary_results_path)

        with open(f"{summary_results_path}{self.parameters}_ps{self.session_number}.json", 'w') as file:
            json.dump(results.to_dict(), file, indent=2)

    def compile_roi_stats(self, roi_temp_store, roi_results):
        warnings.filterwarnings('ignore')  # Ignore warnings that indicate an ROI has only nan values

        axis = 1
        read_start = 0
        write_start = 0
        write_end = -1

        # First loop calculates summary stats for normal ROIs, second loop calculates stats for overall ROI
        for i in range(2):
            roi_results[0, write_start:write_end] = np.count_nonzero(~np.isnan(roi_temp_store[read_start:, :]),
                                                                     axis=axis)  # Count number of non-nan voxels
            roi_results[1, write_start:write_end] = np.nanmean(roi_temp_store[read_start:, :], axis=axis)
            roi_results[2, write_start:write_end] = np.nanstd(roi_temp_store[read_start:, :], axis=axis, ddof=1)
            # Confidence interval calculation
            roi_results[3, write_start:write_end] = Environment_Setup.conf_level_list[int(config.conf_level_number)][1] \
                                                    * roi_results[2, write_start:write_end] \
                                                    / np.sqrt(roi_results[0, write_start:write_end])
            roi_results[4, write_start:write_end] = np.nanmedian(roi_temp_store[read_start:, :], axis=axis)
            roi_results[5, write_start:write_end] = np.nanmin(roi_temp_store[read_start:, :], axis=axis)
            roi_results[6, write_start:write_end] = np.nanmax(roi_temp_store[read_start:, :], axis=axis)

            axis = None
            read_start = 1
            write_start = -1
            write_end = None

        roi_results[7, -1] = np.sum(roi_results[7, 1:-1])  # Calculate excluded voxels

        # Convert ROIs with no voxels from columns with NaNs to zeros
        for column, voxel_num in enumerate(roi_results[0]):
            if voxel_num == 0.0:
                for row in list(range(1, 6)):
                    roi_results[row][column] = 0.0

        warnings.filterwarnings('default')  # Reactivate warnings

        return roi_results

    def noise_correct_data_and_create_excluded_voxel_maps(self, roi_results, roi_temp_store, volumes):
        header, statmap_shape, save_location, stage, excluded_voxels_file_location = self.create_excluded_rois_volume(
            volumes)

        if config.verbose:
            print_outlier_removal_methods(config, self.no_ext_brain)

        # Remove outliers from ROIs
        if config.noise_cutoff:
            roi_results, roi_temp_store, self.noise_threshold = self.noise_threshold_outlier_detection(roi_results,
                                                                                                       roi_temp_store)

            excluded_voxels_file_location = f"{save_location}ExcludedVoxStage{stage}_{self.no_ext_brain}_noiseThreshOutliers.nii.gz"

            create_no_roi_volume(roi_temp_store,
                                 excluded_voxels_file_location,
                                 statmap_shape,
                                 header)

            stage += 1

        if config.gaussian_outlier_detection:
            roi_results, roi_temp_store, self.lower_gaussian_threshold, self.upper_gaussian_threshold = gaussian_outlier_detection(
                roi_results,
                roi_temp_store,
                config
            )

            excluded_voxels_file_location = f"{save_location}ExcludedVoxStage{stage}_{self.no_ext_brain}_gaussianThreshOutliers.nii.gz"

            create_no_roi_volume(roi_temp_store,
                                 excluded_voxels_file_location,
                                 statmap_shape,
                                 header)

            stage += 1

        roi_temp_store, roi_results = self.create_final_atlas_mapping(excluded_voxels_file_location,
                                                                      roi_results,
                                                                      save_location)

        return roi_results, roi_temp_store

    def create_final_atlas_mapping(self, final_excluded_voxels_volume, roi_results, output_folder):
        # Make copy of fMRI volume
        shutil.copy(self.brain, f"{output_folder}/fMRI_volume.nii.gz")

        # Make copy of mni_to_fmri brain
        orig_mni_loc = f"{output_folder}/orig_mni_to_{self.no_ext_brain}.nii.gz"
        shutil.copy(self.mni_brain, orig_mni_loc)

        # Create binary mask using -binv
        binary_mask = fsl_functions(self, output_folder, self.no_ext_brain, 'maths.UnaryMaths',
                                    final_excluded_voxels_volume, 'binary_mask_', 'binarise_and_invert')

        # Fill holes in binary mask using -fillh
        binary_mask_filled = fsl_functions(self, output_folder, self.no_ext_brain, 'maths.UnaryMaths',
                                           binary_mask, 'binary_mask_filled_', 'fill_holes')

        # Multiply mni_to_fMRI with binary_mask_filled
        fsl_functions(self, output_folder, self.no_ext_brain, 'maths.BinaryMaths',
                      self.mni_brain, 'final_mni_to_', binary_mask_filled, 'mul', 'Save output to self')

        # Calculate how many voxels have been filled in for each ROI. This will be subtracted from the excluded voxels
        # row of roi_results to create an accurate count after rerunning the analysis
        roi_results = self.correct_excluded_voxel_amount_by_number_of_filled_voxels(binary_mask,
                                                                                    binary_mask_filled,
                                                                                    orig_mni_loc,
                                                                                    roi_results,
                                                                                    output_folder)

        # Load brains and pre-initialise arrays
        roi_temp_store, _, idxMNI, idxBrain = self.roi_stats_setup()

        # The magic loop
        for counter, roi in enumerate(idxMNI):
            roi_temp_store[int(roi), counter] = idxBrain[counter]

        return roi_temp_store, roi_results

    def correct_excluded_voxel_amount_by_number_of_filled_voxels(self, binary_mask, binary_mask_filled, mni_brain,
                                                                 roi_results, output_folder):
        # binary_mask_filled - binary_mask
        filled_voxel_volume = fsl_functions(self, output_folder, self.no_ext_brain, 'maths.BinaryMaths',
                                            binary_mask_filled, 'filled_voxels_', binary_mask, 'sub')

        filled_voxel_volume = nib.load(filled_voxel_volume)
        filled_voxel_volume = filled_voxel_volume.get_fdata()
        idxBrain = filled_voxel_volume.flatten()

        # Load atlas brain (which has been converted into native space)
        mni_brain = nib.load(mni_brain)
        mni_brain = mni_brain.get_fdata()
        idxMNI = mni_brain.flatten()

        # Find the number of unique ROIs in the atlas
        roiList = list(range(0, len(self._labelArray) - 1))
        roiNum = np.size(roiList)

        # Create arrays to store the values before and after statistics
        filled_voxels_per_roi = np.full([1, roiNum], 0)

        for counter, roi in enumerate(idxMNI):
            if idxBrain[counter] == 1.0:  # Equal to 1.0 if the voxel has been filled with -fillh
                filled_voxels_per_roi[0, int(roi)] += 1

        roi_results[-1, 1:-1] = roi_results[-1, 1:-1] - filled_voxels_per_roi[:, 1:]

        return roi_results

    def create_excluded_rois_volume(self, volumes):
        excluded_vox_save_location = f"{self.save_location}/Excluded_voxels/{self.no_ext_brain}/"
        Utils.check_and_make_dir(f"{self.save_location}/Excluded_voxels/")
        Utils.check_and_make_dir(excluded_vox_save_location)

        statmap, header = Utils.load_brain(self.stat_brain)
        statmap_shape = statmap.shape

        stage = 1

        for file_name, volume in volumes.items():
            file_location = f"{excluded_vox_save_location}/ExcludedVoxStage{stage}_{self.no_ext_brain}_{file_name}.nii.gz"

            create_no_roi_volume(
                volume,
                file_location,
                statmap_shape,
                header
            )

            stage += 1

        return header, statmap_shape, excluded_vox_save_location, stage, file_location

    def noise_threshold_outlier_detection(self, roi_results, roi_temp_store):
        # Invert binary mask using -binv
        binary_mask = fsl_functions(self, self.save_location, self.no_ext_brain, 'maths.UnaryMaths',
                                    f'{self.save_location}/bet_{self.no_ext_brain}_mask.nii.gz',
                                    'inverted_bet_mask_', 'binarise_and_invert', 'Save to file list')

        # Multiply mni_to_fMRI with binary_mask_filled
        extra_cranial_voxels = fsl_functions(self, self.save_location, self.no_ext_brain, 'maths.BinaryMaths',
                                             self.stat_brain, 'extra_cranial_voxels_', binary_mask, 'mul',
                                             'Save to file list')

        extra_cranial_vox_volume, _ = Utils.load_brain(extra_cranial_voxels)

        # Calculate noise threshold
        noise_threshold = np.nansum(extra_cranial_vox_volume) / np.count_nonzero(extra_cranial_vox_volume)

        outlier_bool_array = roi_temp_store[1:, :] < noise_threshold
        roi_results = calculate_number_of_outliers_per_roi(outlier_bool_array, roi_results)
        roi_temp_store = remove_outliers(outlier_bool_array, roi_temp_store)

        return roi_results, roi_temp_store, noise_threshold


class MatchedBrain:
    label_array = []
    save_location = None

    def __init__(self, brain_files, parameters):
        self.brains = brain_files
        self.parameters = "_".join([f"{param_name}{param_val}" for param_name, param_val
                                    in zip(config.parameter_dict2, parameters)])

        self.participant_grouped_summarised_results = {participant: [] for participant in brain_files.keys()}
        self.ungrouped_summarised_results = []
        self.ungrouped_raw_results = []

        self.session_averaged_results = []
        self.participant_averaged_results = []

        self.excluded_voxels = None
        self.conf_int = Environment_Setup.conf_level_list[int(config.conf_level_number)]

        # Instance variables to be accessed by participant averaged function
        self.session_number = None
        self.maximum_result = None
        self.minimum_result = None
        self.total_voxels = None
        self.average_voxels = None

        # Copying class attributes here is a workaround for dill, which can't access class attributes.
        self.label_array = self.label_array
        self.save_location = self.save_location

    @classmethod
    def setup_class(cls, participant_list):
        matched_brains = cls.find_shared_params(participant_list)  # Find brains which share parameter combinations

        cls.label_array = Environment_Setup.label_array
        cls.save_location = f"{Environment_Setup.save_location}Overall/"

        Utils.check_and_make_dir(cls.save_location)
        Utils.check_and_make_dir(f"{cls.save_location}/Summarised_results/")
        Utils.check_and_make_dir(f"{cls.save_location}/Summarised_results/Session_averaged_results")
        Utils.check_and_make_dir(f"{cls.save_location}/Summarised_results/Participant_averaged_results")

        matched_brain_list = set()
        for param_combination in matched_brains:
            # Initialise matched brains
            matched_brain_list.add(MatchedBrain(
                matched_brains[param_combination],
                param_combination
            ))

        cls.assign_parameters_to_brains(matched_brain_list, participant_list)

        return matched_brain_list

    @classmethod
    def find_shared_params(cls, participant_list):
        table, _ = Utils.load_paramValues_file()

        ignore_column_loc, critical_column_locs, _ = Utils.find_column_locs(table)

        matched_brains = dict()
        for row in table.itertuples(index=False):
            if ignore_column_loc and str(row[ignore_column_loc]).strip().lower() in (
                    'yes', 'y', 'true'):  # If column is set to ignore then do not include it in analysis
                continue

            elif critical_column_locs:
                values = tuple(map(row.__getitem__, critical_column_locs))

            else:
                values = tuple(['results'])

            if values not in matched_brains.keys():
                matched_brains[values] = dict()

            participant, brain_file = cls.find_brain_object(row, participant_list)

            if participant in matched_brains[values]:
                matched_brains[values][participant].append(brain_file)
            else:
                matched_brains[values][participant] = [brain_file]

        return matched_brains

    @staticmethod
    def find_brain_object(row, participant_list):
        participant = next((participant for participant in participant_list if participant.participant_name == row[0]),
                           None)

        if not participant:
            raise FileNotFoundError(f'Subject {row[0]} exists in {config.parameter_file} but no longer '
                                    f'exists in input folder "{config.input_folder_name}".\nRemove these rows from the '
                                    f'{config.parameter_file} file, or re-add them into the input folder.')

        brain = next((brain for brain in participant.brains if brain.no_ext_brain == row[1]), None)

        if not brain:
            raise FileNotFoundError(f'{row[1]} for subject {row[0]} exists in {config.parameter_file} but no longer '
                                    f'exists in input folder "{config.input_folder_name}".\nRemove these rows from the '
                                    f'{config.parameter_file} file, or re-add them into the input folder.')

        return participant.participant_name, brain.no_ext_brain

    def compile_results(self, path, config):
        if config.verbose:
            print(f'Combining results for parameter combination: {self.parameters}')

        self.save_location = f"{path}/{self.save_location}"

        # Load roi_temp_store results from binary numpy file and combine them
        raw_results = []
        for roi_temp_store in self.ungrouped_raw_results:
            raw_results.append(np.load(roi_temp_store))
        self.ungrouped_raw_results = np.concatenate(raw_results, axis=1)

        self.create_array_to_calculate_excluded_voxels()

        self.calculate_and_save_session_averaged_results()
        self.calculate_and_save_participant_averaged_results()

        # Save results
        reformat_and_save_raw_data(self.ungrouped_raw_results, self.label_array,
                                   self.save_location, self.parameters)

        del self.ungrouped_raw_results  # Clear numpy array from memory

        return self

    def create_array_to_calculate_excluded_voxels(self):
        # TODO is this still necessary?
        self.excluded_voxels = copy.deepcopy(self.ungrouped_summarised_results)

        for result in self.excluded_voxels:
            result[0:-1, :] = 0  # Reset roi_results and only retain excluded voxels

        # Collapse list to get total excluded voxels
        self.excluded_voxels = np.sum(self.excluded_voxels, axis=0)

    def calculate_and_save_participant_averaged_results(self):
        for participant in self.participant_grouped_summarised_results:
            for session in self.participant_grouped_summarised_results[participant]:
                for row in session:
                    row[row == 0] = np.nan

        row_labels = ['Total voxels',
                      'Excluded voxels',
                      'Average voxels per session',
                      'Mean',
                      'Std_dev',
                      f'Conf_Int_{self.conf_int[0]}',
                      'Median',
                      'Minimum',
                      'Maximum',
                      'Participants',
                      'Sessions'
                      ]

        results = np.full((len(row_labels), len(self.label_array)), np.inf)

        results[0, :] = self.total_voxels
        results[1, :] = self.excluded_voxels[7]
        results[2, :] = self.average_voxels
        results[7, :] = self.minimum_result
        results[8, :] = self.maximum_result
        results[10, :] = self.session_number

        # Ignore runtime warnings about "Mean of empty slice" or "Degrees of freedom <= 0 for slice"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)

            participant_mean = []
            participant_median = []
            for participant in self.participant_grouped_summarised_results:
                participant_mean.append(np.nanmean(self.participant_grouped_summarised_results[participant], axis=0)[1])
                participant_median.append(
                    np.nanmedian(self.participant_grouped_summarised_results[participant], axis=0)[4])

            results[3, :] = np.nanmean(participant_mean, axis=0)
            results[4, :] = np.nanstd(participant_mean, axis=0)
            results[6, :] = np.nanmedian(participant_median, axis=0)

            # Counting nonzero as sometimes an ROI won't show up for a particular participant
            results[9, :] = np.count_nonzero(~np.isnan(participant_mean), axis=0)  # Participants
            results[5, :] = self.conf_int[1] * results[4, :] / np.sqrt(results[9, :])  # Confidence interval

        results_df = pd.DataFrame(index=pd.Index(row_labels), data=results, columns=self.label_array)

        with open(f"{self.save_location}/Summarised_results/Participant_averaged_results/{self.parameters}.json",
                  'w') as file:
            json.dump(results_df.to_dict(), file, indent=2)

        self.participant_averaged_results = results_df.to_numpy()

        return results_df

    def calculate_and_save_session_averaged_results(self):
        nan_data = copy.deepcopy(self.ungrouped_summarised_results)

        for session in nan_data:
            for row in session:
                row[row == 0] = np.nan

        row_labels = ['Total voxels',
                      'Excluded voxels',
                      'Average voxels per session',
                      'Mean',
                      'Std_dev',
                      f'Conf_Int_{self.conf_int[0]}',
                      'Median',
                      'Minimum',
                      'Maximum',
                      'Sessions']

        results = np.full((len(row_labels), len(self.label_array)), np.inf)

        self.total_voxels = results[0, :] = np.nansum(self.ungrouped_summarised_results, axis=0)[0]  # Total voxels
        results[1, :] = self.excluded_voxels[7]  # Excluded voxels

        # Ignore runtime warnings about "Mean of empty slice" or "Degrees of freedom <= 0 for slice"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)

            self.average_voxels = results[2, :] = np.nanmean(nan_data, axis=0)[0]  # Average voxels
            results[3, :] = np.nanmean(nan_data, axis=0)[1]  # Mean
            results[4, :] = np.nanstd(nan_data, axis=0)[1]  # Standard deviation
            results[6, :] = np.nanmedian(nan_data, axis=0)[4]  # Median

        self.minimum_result = results[7, :] = np.nanmin(self.ungrouped_summarised_results, axis=0)[5]  # Minimum
        self.maximum_result = results[8, :] = np.nanmax(self.ungrouped_summarised_results, axis=0)[6]  # Maximum
        self.session_number = results[9, :] = np.count_nonzero(self.ungrouped_summarised_results, axis=0)[0]  # Sessions

        results[5, :] = self.conf_int[1] * results[4, :] / np.sqrt(results[9, :])  # Confidence interval

        results_df = pd.DataFrame(index=pd.Index(row_labels), data=results, columns=self.label_array)

        with open(f"{self.save_location}/Summarised_results/Session_averaged_results/{self.parameters}.json",
                  'w') as file:
            json.dump(results_df.to_dict(), file, indent=2)

        self.session_averaged_results = results_df.to_numpy()

    @classmethod
    def assign_parameters_to_brains(cls, matched_brains, participant_list):
        brain_list = []
        for participant in participant_list:
            for brain in participant.brains:
                brain_list.append(brain)

        for parameter_comb in matched_brains:
            for brain in brain_list:
                try:
                    if brain.no_ext_brain in parameter_comb.brains[brain.participant_name]:
                        brain.parameters = parameter_comb.parameters

                except KeyError:
                    pass

        # Assign session number to brains if participants have more than one scan with the same parameter combination
        for parameter_comb in matched_brains:
            for subj in parameter_comb.brains:
                if len(parameter_comb.brains[subj]) > 1:
                    session_count = 1

                    for brain in brain_list:
                        try:
                            if brain.participant_name == subj and \
                                    brain.no_ext_brain in parameter_comb.brains[brain.participant_name]:
                                brain.session_number = session_count
                                session_count += 1

                        except KeyError:
                            pass

    def atlas_scale(self, max_roi_stat, brain_number_current, brain_number_total, statistic_num, atlas_path, data,
                    config):
        """Produces up to three scaled NIFTI files. Within brains, between brains (based on rois), between brains
        (based on the highest seen value of all brains and rois)."""

        if data == 'Session averaged':
            results = self.session_averaged_results
            subfolder = 'Session_averaged_results/'
            statistic_labels = config.statistic_options['Session averaged']

        else:
            results = self.participant_averaged_results
            subfolder = 'Participant_averaged_results/'
            statistic_labels = config.statistic_options['Participant averaged']

        if config.verbose and max(max_roi_stat) != 0.0:
            print(f'Creating {statistic_labels[statistic_num]} NIFTI_ROI file for parameter combination '
                  f'{brain_number_current + 1}/{brain_number_total}: {self.parameters}.')

        elif config.verbose \
                and statistic_labels[statistic_num] == 'Excluded_voxels_amount' \
                and max(max_roi_stat) == 0.0:
            print(f'Not creating {statistic_labels[statistic_num]} NIFTI_ROI file for parameter combination '
                  f'{brain_number_current + 1}/{brain_number_total}: {self.parameters} as no voxels have been excluded.')

            return

        self.scale_and_save_atlas_images(atlas_path, max_roi_stat, results, statistic_num,
                                         f"{self.save_location}NIFTI_ROI/{subfolder}",
                                         f"{self.parameters}_{statistic_labels[statistic_num]}")

    @classmethod
    def scale_and_save_atlas_images(cls, atlas_path, max_stat, results, statistic_num, file_path, file_name):
        if type(max_stat) is np.ndarray:
            roi_scaled_stat = [(y / x) * 100 for x, y in zip(max_stat, results[statistic_num, :])]

            # Find maximum statistic value (excluding No ROI and overall category)
            global_scaled_stat = [(y / max(max_stat[1:-1])) * 100 for y in results[statistic_num, :]]
        else:
            roi_scaled_stat, global_scaled_stat = None, None

        atlas = nib.load(atlas_path)
        affine = atlas.affine
        atlas = atlas.get_fdata()

        unscaled_stat, within_roi_stat, mixed_roi_stat = cls.group_roi_stats(atlas, global_scaled_stat,
                                                                             roi_scaled_stat, statistic_num,
                                                                             results)
        # Convert atlas to NIFTI and save it
        scale_stats = [
            (unscaled_stat, f"{file_path}/{file_name}.nii.gz"),
            (within_roi_stat, f"{file_path}/{file_name}_within_roi_scaled.nii.gz"),
            # TODO reimplement mixed_roi_stat
            # (mixed_roi_stat, f"{file_path}/{file_name}_mixed_roi_scaled.nii.gz")
        ]

        for scale_stat in scale_stats:
            if scale_stat[0] is not None:
                scaled_brain = nib.Nifti1Image(np.float64(scale_stat[0]), affine)
                scaled_brain.to_filename(scale_stat[1])

    @staticmethod
    def group_roi_stats(atlas, global_scaled_stat, roi_scaled_stat, statistic_num, results):
        # Iterate through each voxel in the atlas
        atlas = atlas.astype(int)

        # Assign stat values for each ROI all at once
        unscaled_stat = results[statistic_num, atlas]
        # Make ROI group 0 (No ROI) nan so it does not effect colourmap when viewing in fsleyes
        unscaled_stat[atlas == 0] = np.nan

        if global_scaled_stat and roi_scaled_stat:
            within_roi_stat = np.array(roi_scaled_stat)[atlas]
            mixed_roi_stat = np.array(global_scaled_stat)[atlas]

            # Make ROI group 0 (No ROI) nan so it does not effect colourmap when viewing in fsleyes
            within_roi_stat[atlas == 0] = np.nan
            mixed_roi_stat[atlas == 0] = np.nan
        else:
            within_roi_stat = None
            mixed_roi_stat = None

        return unscaled_stat, within_roi_stat, mixed_roi_stat


def create_no_roi_volume(volume, save_location, statmap_shape, header):
    data = volume[0, :].copy()
    data = data.reshape(statmap_shape)

    brain = nib.Nifti1Pair(data, None, header)
    nib.save(brain, save_location)


def print_outlier_removal_methods(config, brain):
    string = []
    if config.noise_cutoff:
        string.append('noise cutoff')
    if config.gaussian_outlier_detection:
        string.append('gaussian outlier detection')

    if string:
        print(f'Running outlier removal using {" & ".join(string)} for volume: {brain}')


def run_flirt_cost_function(fslfunc, ref, init, out_file, matrix_file, config, wmseg=None):
    fslfunc.inputs.reference = ref
    fslfunc.inputs.args = f"-init {init}"  # args used as in_matrix_file method not working
    fslfunc.inputs.out_file = out_file
    fslfunc.inputs.out_matrix_file = matrix_file

    if wmseg:
        fslfunc.inputs.cost = 'bbr'
        fslfunc.inputs.wm_seg = wmseg

    if config.verbose_cmd_line_args:
        print(f"{fslfunc.cmdline}\n")

    output = fslfunc.run()
    cost_func = float(re.search("[0-9]*\.[0-9]+", output.runtime.stdout)[0])

    # Clean up files
    os.remove(out_file)
    os.remove(matrix_file)

    return cost_func


def fsl_functions(obj, save_location, no_ext_brain, func, input_volume, prefix, *argv):
    """Run an FSL function using NiPype."""
    save_override = False
    current_mat = None
    current_brain, fslfunc, suffix = fsl_functions_setup(func, input_volume, no_ext_brain, prefix, save_location)

    # Arguments dependent on FSL function used
    if func == 'MCFLIRT':
        fslfunc.inputs.save_rms = True

    elif func == 'BET':
        fslfunc.inputs.functional = True

    elif func == 'FLIRT':
        fslfunc.inputs.reference = argv[0]
        fslfunc.inputs.dof = config.dof
        current_mat = fslfunc.inputs.out_matrix_file = f'{save_location}{prefix}{no_ext_brain}.mat'

    elif func == 'maths.Threshold':
        fslfunc.inputs.thresh = 0.5

    elif func == 'maths.UnaryMaths':
        if argv[0] == 'binarise_wmseg':
            fslfunc.inputs.operation = 'bin'
            current_brain = fslfunc.inputs.out_file = f"{save_location}{prefix}{no_ext_brain}_fast_wmseg.nii.gz"

        elif argv[0] == 'binarise_and_invert':
            fslfunc.inputs.operation = 'binv'

            try:
                if argv[1] == 'Save to file list':
                    save_override = True
            except IndexError:
                pass

        elif argv[0] == 'fill_holes':
            fslfunc.inputs.operation = 'fillh'

    elif func == 'maths.BinaryMaths':
        fslfunc.inputs.operand_file = argv[0]
        fslfunc.inputs.operation = argv[1]

        try:
            if argv[2] == 'Save output to self':
                obj.mni_brain = current_brain
            elif argv[2] == 'Save to file list':
                save_override = True
        except IndexError:
            pass

    elif func == 'EpiReg':
        fslfunc.inputs.epi = input_volume
        fslfunc.inputs.t1_brain = argv[0]
        fslfunc.inputs.t1_head = argv[1]
        fslfunc.inputs.wmseg = argv[2]

        # Gets around a bug in nipype code that only allows the wmseg value to be the outbase value with _fast_wmseg suffix
        fslfunc.inputs.out_base = argv[2].replace('_fast_wmseg.nii.gz', '')
        current_mat = f'{fslfunc.inputs.out_base}.mat'
        current_brain = f'{fslfunc.inputs.out_base}.nii.gz'

    elif func == 'FAST':
        fslfunc.inputs.in_files = input_volume
        fslfunc.inputs.out_basename = save_location

    elif func == 'ConvertXFM':
        if len(argv) > 0 and argv[0] == 'concat_xfm':
            fslfunc.inputs.in_file2 = obj.anat_to_mni_mat
            fslfunc.inputs.concat_xfm = True
        else:
            fslfunc.inputs.invert_xfm = True

    elif func == 'ApplyXFM':
        fslfunc.inputs.apply_xfm = True
        fslfunc.inputs.in_matrix_file = argv[0]
        fslfunc.inputs.reference = argv[1]
        fslfunc.inputs.interp = argv[2]
        current_mat = fslfunc.inputs.out_matrix_file = f"{save_location}{prefix}{no_ext_brain}.mat"

    if config.verbose_cmd_line_args:
        print(f"{fslfunc.cmdline}\n")

    fslfunc.run()

    if func not in ('maths.Threshold', 'maths.UnaryMaths', 'maths.BinaryMaths', 'FAST') or save_override:
        fsl_function_file_handle(current_brain, current_mat, func, no_ext_brain, obj, prefix, save_location, suffix)

    if func == 'FLIRT':
        return current_mat
    elif func == 'EpiReg':
        return f'{save_location}{prefix}{no_ext_brain}.mat'
    elif func == 'FAST':
        return

    return current_brain


def fsl_function_file_handle(current_brain, current_mat, func, no_ext_brain, obj, prefix, save_location, suffix):
    if func in ('FLIRT', 'ApplyXFM'):
        obj.file_list.extend([current_brain, current_mat])

    elif func == 'EpiReg':
        new_base = f'{save_location}{prefix}{no_ext_brain}'
        renamed_brain = f'{new_base}.nii.gz'
        renamed_mat = f'{new_base}.mat'
        renamed_wmedge = f'{new_base}_fast_wmedge.nii.gz'
        renamed_wmseg = f'{new_base}_fast_wmseg.nii.gz'

        os.rename(current_brain, renamed_brain)
        os.rename(current_mat, renamed_mat)
        os.rename(f'{save_location}{no_ext_brain}_fast_wmedge.nii.gz', renamed_wmedge)
        os.rename(f'{save_location}{no_ext_brain}_fast_wmseg.nii.gz', renamed_wmseg)

        obj.file_list.extend([renamed_brain, renamed_mat, renamed_wmedge, renamed_wmseg])

    elif func == 'BET':
        obj.file_list.extend([current_brain, f"{save_location}{prefix}{no_ext_brain}_mask{suffix}"])

    elif func == 'MCFLIRT':
        # Find all the motion correction files that are not the actual brain volume
        mc_files = [direc for direc in os.listdir(f"{save_location}motion_correction_files/")
                    if re.search(f'^{prefix}{no_ext_brain}{suffix}(?!$)', direc)]

        # Remove .nii.gz from middle of string
        for file in mc_files:
            os.rename(f"{save_location}motion_correction_files/{file}",
                      f"{save_location}motion_correction_files/{file.replace(suffix, '')}")
    else:
        obj.file_list.append(current_brain)


def fsl_functions_setup(func, input, no_ext_brain, prefix, save_location):
    if 'maths' in func:
        func = func.split('.')[1]
        fslfunc = getattr(maths, func)()
    else:
        fslfunc = getattr(fsl, func)()

    # Standard variables that may be changed for specific FSL functions
    suffix = '.nii.gz'
    fslfunc.inputs.output_type = 'NIFTI_GZ'

    current_brain = None
    if func not in ('EpiReg', 'FAST'):
        fslfunc.inputs.in_file = input
        current_brain = fslfunc.inputs.out_file = f"{save_location}{prefix}{no_ext_brain}{suffix}"

    if func == 'ConvertXFM':
        suffix = '.mat'

    elif func == 'MCFLIRT':
        Utils.check_and_make_dir(f"{save_location}motion_correction_files/")
        current_brain = fslfunc.inputs.out_file = f"{save_location}motion_correction_files/{prefix}{no_ext_brain}{suffix}"

    return current_brain, fslfunc, suffix


def calculate_confidence_interval(data, alpha, roi=None):
    warnings.filterwarnings(action='ignore',
                            category=PendingDeprecationWarning)  # Silences a deprecation warning from bootstrapping library using outdated numpy matrix instead of numpy array

    if roi is None:
        data = data.flatten()
        values = np.array([x for x in data if str(x) != 'nan'])
    else:
        values = np.array([x for x in data[roi, :] if str(x) != 'nan'])

    results = bs.bootstrap(values, stat_func=bs_stats.mean, alpha=alpha, iteration_batch_size=10, num_threads=-1)
    conf_int = (results.upper_bound - results.lower_bound) / 2
    # TODO: Change this so it returns asymmetric bootstrap and implement bootstrapping for pooled analysis

    warnings.simplefilter(action='default', category=PendingDeprecationWarning)

    return results.value, conf_int


def roi_stats_bootstrap(roi_temp_store, roi_results, roiNum, brain_number_current, brain_number_total):
    for counter, roi in enumerate(list(range(0, roiNum + 1))):
        if config.verbose:
            print(f"  - Bootstrapping ROI {counter + 1}/{roiNum + 1} "
                  f"for fMRI volume {brain_number_current + 1}/{brain_number_total}.")

        if counter < roiNum:
            roi_results[1, roi], roi_results[3, roi] = calculate_confidence_interval(roi_temp_store,
                                                                                     config.bootstrap_alpha,
                                                                                     roi=roi)
        else:
            # Calculate overall statistics
            roi_results[1, -1], roi_results[3, -1] = calculate_confidence_interval(roi_temp_store[1:, :],
                                                                                   config.bootstrap_alpha)
    return roi_results


def reformat_and_save_raw_data(roi_temp_store, labelArray, save_location, no_ext_brain, session_number=None):
    if session_number is None:
        session_suffix = ''
    else:
        session_suffix = f'_ps{session_number}'

    formatted_roi_temp_store = roi_temp_store.copy()  # Copy roi_temp_store so changes do get saved later

    # Reorganise matrix to later remove nan rows
    formatted_roi_temp_store = formatted_roi_temp_store.transpose()
    i = np.arange(formatted_roi_temp_store.shape[1])
    # Find indices of nans and put them at the end of each column
    a = np.isnan(formatted_roi_temp_store).argsort(0, kind='mergesort')
    # Reorganise matrix with nans at end
    formatted_roi_temp_store[:] = formatted_roi_temp_store[a, i]

    raw_results = pd.DataFrame(data=formatted_roi_temp_store,
                               columns=labelArray[:-1])

    # Remove the required rows from the dataframe
    raw_results = raw_results.drop(raw_results.columns[0], axis=1)

    # Remove rows where all columns have NaNs (essential to keep file size down)
    raw_results = raw_results.dropna(axis=0, how='all')

    # Convert to dict and get rid of row numbers to significantly decrease file size
    roidict = Utils.dataframe_to_dict(raw_results)

    raw_results_path = f"{save_location}Raw_results/"

    Utils.check_and_make_dir(raw_results_path)
    Utils.check_and_make_dir(f"{raw_results_path}/numpy_arrays/", delete_old=False)

    with open(f"{raw_results_path}{no_ext_brain}{session_suffix}_raw.json", 'w') as file:
        json.dump(roidict, file, indent=2)

    # Save roi_temp_store as binary numpy file to preserve space
    numpy_array_loc = f"{raw_results_path}/numpy_arrays/{no_ext_brain}{session_suffix}_raw.npy"
    np.save(numpy_array_loc, roi_temp_store)

    return numpy_array_loc


def verify_param_values():
    """Compare critical parameter choices to those in paramValues.csv. Exit with exception if discrepancy found."""
    table, _ = Utils.load_paramValues_file()
    Utils.find_column_locs(table)


def construct_combined_results(directory, analysis_type='overall', subfolder=''):
    if subfolder == 'session averaged':
        directory = f"{directory}/Summarised_results/Session_averaged_results"
    elif subfolder == 'participant averaged':
        directory = f"{directory}/Summarised_results/Participant_averaged_results/"
    else:
        directory = f"{directory}/Summarised_results/"

    json_file_list = [os.path.basename(f) for f in glob(f"{directory}/*.json")]

    save_combined_results_file(directory, json_file_list)

    if analysis_type == 'participant':
        save_averaged_results_file(directory, json_file_list)


def save_averaged_results_file(directory, json_file_list):
    Utils.check_and_make_dir(f"{directory}/Averaged_results/")

    parameter_dict = {}
    for jsn in json_file_list:
        if 'combined_results' in jsn:
            continue

        param, _, _ = jsn.partition('_ps')

        if param not in parameter_dict:
            parameter_dict[param] = [jsn]
        else:
            parameter_dict[param].append(jsn)

    overall_averaged_dataframe = pd.DataFrame()
    for parameter, jsns in parameter_dict.items():
        session_averaged_dataframe = pd.DataFrame()

        for jsn in jsns:
            with open(f'{directory}/{jsn}', "r") as results:
                results = pd.DataFrame(json.load(results))
                row_order = results.index

            session_averaged_dataframe = pd.concat((session_averaged_dataframe, results))

        # Get the summarised results across all json files for each parameter combination
        grouped_by_stat = session_averaged_dataframe.groupby(session_averaged_dataframe.index)
        averaged_dataframe = grouped_by_stat.mean().reindex(row_order)

        overall_averaged_dataframe = pd.concat((overall_averaged_dataframe, averaged_dataframe))

        with open(f"{directory}/Averaged_results/{parameter}.json", 'w') as file:
            json.dump(averaged_dataframe.to_dict(), file, indent=2)

    # Repeat the above to create the overall summarised results
    grouped_by_stat = overall_averaged_dataframe.groupby(overall_averaged_dataframe.index)
    averaged_dataframe = grouped_by_stat.mean().reindex(row_order)

    with open(f"{directory}/Averaged_results/overall.json", 'w') as file:
        json.dump(averaged_dataframe.to_dict(), file, indent=2)


def save_combined_results_file(directory, json_file_list):
    session_dict = split_jsons_by_session(json_file_list)

    for session, jsns in session_dict.items():
        if not jsns:
            continue

        combined_dataframe = pd.DataFrame()

        for jsn in jsns:
            if 'combined_results' in jsn:
                continue

            # Splits a file name. For example from hb1_ip2.json into ['hb1', 'ip2']
            parameters = list(filter(None, re.split(f'_|ps[0-9]*|.json', jsn)))

            # Remove critical parameter name from file name. For example turn ['hb1', 'ip2'] into [1, 2]
            for counter, critical_parameter in enumerate(config.parameter_dict2):
                parameters[counter] = parameters[counter].replace(critical_parameter, '')

            current_dataframe = pd.read_json(f"{directory}/{jsn}")
            current_dataframe = current_dataframe.transpose()

            for counter, parameter_name in enumerate(config.parameter_dict):
                current_dataframe[parameter_name] = parameters[counter]  # Add parameter columns

            current_dataframe['File_name'] = Utils.strip_ext(jsn)

            if combined_dataframe.empty:
                combined_dataframe = current_dataframe

            else:
                combined_dataframe = pd.concat([combined_dataframe, current_dataframe], sort=True)

        if session == 0:
            session_suffix = ''
        else:
            session_suffix = f'_ps{session}'

        # Save combined results
        combined_dataframe = combined_dataframe.reset_index()
        combined_dataframe.to_json(f"{directory}/combined_results{session_suffix}.json", orient='records', indent=2)


def split_jsons_by_session(json_file_list):
    session_dict = {0: []}

    for jsn in json_file_list:
        if '_ps' in jsn:
            session_number = re.findall('_ps[0-9]*', jsn)[0].split('_ps')[-1]

            if session_number not in session_dict:
                session_dict[session_number] = []

            session_dict[session_number].append(jsn)

        else:
            session_dict[0].append(jsn)

    return session_dict


def gaussian_outlier_detection(roi_results, roi_temp_store, config):
    # Convert gaussian outlier contamination into confidence interval to retain for scipy norm.interval function
    confidence_interval = 1 - config.gaussian_outlier_contamination

    # Fit a gaussian to the data using EllipticEnvelope
    outliers, lower_gaussian_threshold, upper_gaussian_threshold = outlier_detection_using_gaussian(
        data=roi_temp_store[1:, :],
        contamination=confidence_interval,
        config=config
    )

    outliers.sort()  # Sort to make it easier to find start and end

    outlier_bool_array = np.isin(roi_temp_store[1:, :], outliers)
    roi_results = calculate_number_of_outliers_per_roi(outlier_bool_array, roi_results)
    roi_temp_store = remove_outliers(outlier_bool_array, roi_temp_store)

    return roi_results, roi_temp_store, lower_gaussian_threshold, upper_gaussian_threshold


def outlier_detection_using_gaussian(data, contamination, config):
    values = data.copy()
    values = values[~np.isnan(values)]

    prediction = norm(*norm.fit(values))
    gaussian_lims = prediction.interval(contamination)

    lower_gaussian_threshold = None
    upper_gaussian_threshold = None

    if config.gaussian_outlier_location == 'below gaussian':
        outliers = [value for value in values if value < gaussian_lims[0]]
        lower_gaussian_threshold = gaussian_lims[0]

    elif config.gaussian_outlier_location == 'above gaussian':
        outliers = [value for value in values if value > gaussian_lims[1]]
        upper_gaussian_threshold = gaussian_lims[1]

    elif config.gaussian_outlier_location == 'both':
        outliers = [value for value in values if value < gaussian_lims[0] or value > gaussian_lims[1]]
        lower_gaussian_threshold = gaussian_lims[0]
        upper_gaussian_threshold = gaussian_lims[1]

    else:
        raise Exception('gaussian_outlier_location not valid')

    return outliers, lower_gaussian_threshold, upper_gaussian_threshold


def calculate_number_of_outliers_per_roi(outlier_bool_array, roi_results):
    # Count how many voxels are below noise threshold for each ROI
    below_threshold = np.count_nonzero(outlier_bool_array, axis=1)
    # Assign voxels below threshold to excluded voxels column
    roi_results[7, 1:-1] += below_threshold

    return roi_results


def remove_outliers(outlier_bool_array, roi_temp_store):
    # Convert from bool to float64 to be able to assign values
    outlier_float_array = np.max(outlier_bool_array, axis=0).astype('float64')

    # Replace False with nans and True with statistic value
    outlier_float_array[outlier_float_array == 0.0] = np.nan

    # Use the transpose of roi_temp_store and outlier bool array to make sure output is ordered by columns not rows
    outlier_float_array[outlier_float_array == 1.0] = roi_temp_store[1:, :].T[outlier_bool_array.T]

    # Set any outlier ROI values to nan
    roi_temp_store[1:, :][outlier_bool_array] = np.nan

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', r'All-NaN slice encountered')
        # Stack no ROI values on top of array containing outlier values, and then use nan max to condense into one array
        roi_temp_store[0, :] = np.nanmax(np.vstack((outlier_float_array, roi_temp_store[0, :])), axis=0)

    return roi_temp_store



import json

from .utils import *


def printResults(config_name):
    print('--- Printing results ---')
    print('Select the results directory created by fRAT.')
    result_loc = Utils.file_browser(title='Select the directory output by the fRAT')

    config = Utils.load_config(f'{Path(os.path.abspath(__file__)).parents[1]}/configuration_profiles/roi_analysis',
                               config_name)

    if config.averaging_type == 'Session averaged':
        subfolder = 'Session_averaged_results'
    else:
        subfolder = 'Participant_averaged_results'

    print(f'Showing {config.averaging_type} results.\n')

    with open(f"{result_loc}/Overall/Summarised_results/{subfolder}/combined_results.json", "r") as results:
        results = json.load(results)
        rois = sorted({result['index'] for result in results})  # Using set returns only unique values

    blacklist = ['index', 'File_name', *config.parameter_dict1]

    chosen_rois = user_input(rois)

    for roi in chosen_rois:  # For each chosen roi
        print(f"\n----------------------Chosen ROI {roi}: {rois[roi]}----------------------------")
        for result in results:  # For each entry_create in the combined_df
            if result['index'] == rois[roi]:  # If the entry_create is the correct roi, print the result
                print(f"\n -- File name: {result['File_name']}")
                for param in config.parameter_dict1:
                    print(f" -- {param}: {result[param]}")

                [print(f"{key}: {value}") for key, value in result.items() if key not in blacklist]
        print(f"----------------------Chosen ROI {roi}: {rois[roi]} end--------------------------")


def user_input(rois):
    print(f"--- Analysed ROIs ---")

    for counter, roi in enumerate(rois):
        print(f"{counter}: {roi}")

    while True:
        roi_ans = input(
        "Type a comma-separated list of the ROIs (listed above) you want to produce a figure for, "
        "'e.g. 2, 15, 7, 23' or 'all' for all rois. \nAlternatively press enter to skip this step: ")

        if roi_ans.lower() == "all":
            chosen_rois = list(range(0, len(rois)))

        elif len(roi_ans) > 0:
            chosen_rois = [x.strip() for x in roi_ans.split(',')]  # Split by comma and whitespace

            try:
                chosen_rois = list(map(int, chosen_rois))  # Convert each list item to integers
            except ValueError:
                print('Comma-separated list contains non integers.\n')
                chosen_rois = []

            # Error checking for list indices out of range
            for roi in chosen_rois:
                if roi > len(rois) - 1:
                    print('List contains non-valid selections.\n')
                    chosen_rois = False
                    break

            if not chosen_rois:
                continue

        else:  # Else statement for blank input, this skips printing results
            chosen_rois = []
            break

        if isinstance(chosen_rois, int):
            chosen_rois = [chosen_rois]

        return chosen_rois


if __name__ == '__main__':
    printResults()
